import datetime

from fme.core.coordinates import HorizontalCoordinates, VerticalCoordinate
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.device import get_device
from fme.core.mask_provider import MaskProvider


class DatasetProperties:
    def __init__(
        self,
        variable_metadata: dict[str, VariableMetadata],
        vertical_coordinate: VerticalCoordinate,
        horizontal_coordinates: HorizontalCoordinates,
        mask_provider: MaskProvider,
        timestep: datetime.timedelta,
        is_remote: bool,
    ):
        self.variable_metadata = variable_metadata
        self.vertical_coordinate = vertical_coordinate
        self.horizontal_coordinates = horizontal_coordinates
        self.mask_provider = mask_provider
        self.timestep = timestep
        self.is_remote = is_remote

    def to_device(self) -> "DatasetProperties":
        device = get_device()
        return DatasetProperties(
            self.variable_metadata,
            self.vertical_coordinate.to(device),
            self.horizontal_coordinates.to(device),
            self.mask_provider.to(device),
            self.timestep,
            self.is_remote,
        )

    def update(self, other: "DatasetProperties"):
        self.is_remote = self.is_remote or other.is_remote
        if self.timestep != other.timestep:
            raise ValueError("Inconsistent timesteps between datasets")
        if self.variable_metadata != other.variable_metadata:
            raise ValueError("Inconsistent metadata between datasets")
        if self.vertical_coordinate != other.vertical_coordinate:
            raise ValueError("Inconsistent vertical coordinates between datasets")
        if self.horizontal_coordinates != other.horizontal_coordinates:
            raise ValueError("Inconsistent horizontal coordinates between datasets")
        if self.mask_provider != other.mask_provider:
            raise ValueError("Inconsistent mask providers between datasets")

    def update_merged_dataset(self, other: "DatasetProperties"):
        if isinstance(self.variable_metadata, dict):
            self.variable_metadata.update(other.variable_metadata)
        self.mask_provider.update(other.mask_provider)
        self.is_remote = self.is_remote or other.is_remote
        if self.timestep != other.timestep:
            raise ValueError("Inconsistent timesteps between datasets")
        if self.horizontal_coordinates != other.horizontal_coordinates:
            raise ValueError("Inconsistent horizontal coordinates between datasets")



from collections.abc import Mapping

import torch
import xarray as xr

from fme.core.dataset.data_typing import VariableMetadata
from fme.core.typing_ import TensorMapping
from fme.core.wandb import Image

from ..plotting import plot_paneled_data


class SnapshotAggregator:
    """
    An aggregator that records the first sample of the last batch of data.
    """

    _captions = {
        "full-field": (
            "{name} one step full field for first sample in last batch; "
            "(top) generated and (bottom) target [{units}]"
        ),
        "residual": (
            "{name} one step residual (prediction - previous time) for first sample in "
            "last batch; (top) generated and (bottom) target [{units}]"
        ),
        "error": (
            "{name} one step full field error (generated - target) "
            "for first sample in last batch [{units}]"
        ),
    }

    def __init__(
        self, dims: list[str], metadata: Mapping[str, VariableMetadata] | None = None
    ):
        """
        Args:
            dims: Dimensions of the data.
            metadata: Mapping of variable names their metadata that will
                used in generating logged image captions.
        """
        self._dims = dims
        if metadata is None:
            self._metadata: Mapping[str, VariableMetadata] = {}
        else:
            self._metadata = metadata

    @torch.no_grad()
    def record_batch(
        self,
        loss: float,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
    ):
        self._loss = loss
        self._target_data = target_data
        self._gen_data = gen_data
        self._target_data_norm = target_data_norm
        self._gen_data_norm = gen_data_norm

    def _get_data(self) -> tuple[TensorMapping, TensorMapping, TensorMapping]:
        time_dim = 1
        input_time = 0
        target_time = 1
        gen, target, input = {}, {}, {}
        for name in self._gen_data.keys():
            # use first sample in batch
            gen[name] = (
                self._gen_data[name]
                .select(dim=time_dim, index=target_time)[0]
                .cpu()
                .numpy()
            )
            target[name] = (
                self._target_data[name]
                .select(dim=time_dim, index=target_time)[0]
                .cpu()
                .numpy()
            )
            input[name] = (
                self._target_data[name]
                .select(dim=time_dim, index=input_time)[0]
                .cpu()
                .numpy()
            )
        return gen, target, input

    @torch.no_grad()
    def get_logs(self, label: str) -> dict[str, Image]:
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        image_logs = {}
        gen, target, input = self._get_data()
        for name in gen:
            images = {}
            images["error"] = [[(gen[name] - target[name])]]
            images["full-field"] = [[gen[name]], [target[name]]]
            images["residual"] = [
                [(gen[name] - input[name])],
                [(target[name] - input[name])],
            ]
            for key, data in images.items():
                if key == "error" or key == "residual":
                    diverging = True
                else:
                    diverging = False
                caption = self._get_caption(key, name)
                wandb_image = plot_paneled_data(data, diverging, caption=caption)
                image_logs[f"image-{key}/{name}"] = wandb_image
        image_logs = {f"{label}/{key}": image_logs[key] for key in image_logs}
        return image_logs

    def _get_caption(self, key: str, name: str) -> str:
        if name in self._metadata:
            caption_name = self._metadata[name].long_name
            units = self._metadata[name].units
        else:
            caption_name, units = name, "unknown_units"
        caption = self._captions[key].format(name=caption_name, units=units)
        return caption

    def get_dataset(self) -> xr.Dataset:
        gen, target, input = self._get_data()
        ds = xr.Dataset()
        for name in gen:
            if name in self._metadata:
                long_name = self._metadata[name].long_name
                units = self._metadata[name].units
            else:
                long_name = name
                units = "unknown_units"
            metadata_attrs = {"long_name": long_name, "units": units}
            ds[f"error_map-{name}"] = xr.DataArray(
                data=(gen[name] - target[name]), dims=self._dims, attrs=metadata_attrs
            )
            ds[f"gen_full_field_map-{name}"] = xr.DataArray(
                data=gen[name],
                dims=self._dims,
                attrs=metadata_attrs,
            )
            ds[f"gen_residual_map-{name}"] = xr.DataArray(
                data=gen[name] - input[name],
                dims=self._dims,
                attrs=metadata_attrs,
            )
            ds[f"target_residual_map-{name}"] = xr.DataArray(
                data=target[name] - input[name],
                dims=self._dims,
                attrs=metadata_attrs,
            )
        return ds



from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from matplotlib.colors import Colormap
from matplotlib.figure import Figure

from fme.core.wandb import Image, WandB


def get_cmap_limits(data: np.ndarray, diverging=False) -> tuple[float, float]:
    vmin = np.nanmin(data)
    vmax = np.nanmax(data)
    if diverging:
        vmax = max(abs(vmin), abs(vmax))
        vmin = -vmax
    return vmin, vmax


def plot_imshow(
    data: np.ndarray,
    vmin: float | None = None,
    vmax: float | None = None,
    cmap: str | Colormap | None = None,
    flip_lat: bool = True,
    use_colorbar: bool = True,
) -> Figure:
    """Plot a 2D array using imshow, ensuring figure size is same as array size."""
    min_ = np.nanmin(data) if vmin is None else vmin
    max_ = np.nanmax(data) if vmax is None else vmax
    if len(data.shape) == 3:
        data = fold_healpix_data(data, fill_value=0.5 * (min_ + max_))
    if flip_lat:
        lat_dim = -2
        data = np.flip(data, axis=lat_dim)

    if use_colorbar:
        height, width = data.shape
        colorbar_width = max(1, int(0.025 * width))
        range_ = np.linspace(min_, max_, height)
        range_ = np.repeat(range_[:, np.newaxis], repeats=colorbar_width, axis=1)
        range_ = np.flipud(range_)  # wandb images start from top (and left)
        padding = np.zeros((height, colorbar_width)) + np.nan
        data = np.concatenate((data, padding, range_), axis=1)

    # make figure size (in pixels) be the same as array size
    figsize = np.array(data.T.shape) / plt.rcParams["figure.dpi"]
    fig = Figure(figsize=figsize)  # create directly for cleanup when it leaves scope
    ax = fig.add_axes([0, 0, 1, 1])
    ax.imshow(data, cmap=cmap, vmin=vmin, vmax=vmax)
    ax.set_axis_off()
    return fig


def fold_healpix_data(data: np.ndarray, fill_value: float) -> np.ndarray:
    if data.shape[0] != 12:
        raise ValueError(
            "first dimension must be 12 (face) for healpix data, "
            f"got shape {data.shape}"
        )
    # we want to panel the data like this, numbered by first dimension index
    # -----------------
    # |   |   |   |   |
    # |   |   |   |3  |
    # -----------------
    # |   |   |   |   |
    # |   |   |2  |7  |
    # -----------------
    # |   |   |   |   |
    # |   |1  |6  |10 |
    # -----------------
    # |   |   |   |   |
    # |0  |5  |9  |   |
    # -----------------
    # |   |   |   |   |
    # |4  |8  |   |   |
    # -----------------
    # |   |   |   |   |
    # |11 |   |   |   |
    # -----------------
    blank_panel = np.full_like(data[0], fill_value)
    panels = [
        [blank_panel, blank_panel, blank_panel, data[3]],
        [blank_panel, blank_panel, data[2], data[7]],
        [blank_panel, data[1], data[6], data[10]],
        [data[0], data[5], data[9], blank_panel],
        [data[4], data[8], blank_panel, blank_panel],
        [data[11], blank_panel, blank_panel, blank_panel],
    ]
    return np.concatenate([np.concatenate(row, axis=1) for row in panels], axis=0)


def fold_if_healpix_data(data: np.ndarray, fill_value: float) -> np.ndarray:
    if data.shape[0] == 12:
        return fold_healpix_data(data, fill_value)
    return data


def plot_paneled_data(
    data: list[list[np.ndarray]],
    diverging: bool,
    caption: str | None = None,
) -> Image:
    """Plot a list of 2D data arrays in a paneled plot."""
    if diverging:
        cmap = "RdBu_r"
    else:
        cmap = None
    vmin = np.inf
    vmax = -np.inf
    for row in data:
        for arr in row:
            vmin = min(vmin, np.nanmin(arr))
            vmax = max(vmax, np.nanmax(arr))
    if diverging:
        vmax = max(abs(vmin), abs(vmax))
        vmin = -vmax
    if caption is not None:
        caption += " "
    else:
        caption = ""

    caption += f"vmin={vmin:.4g}, vmax={vmax:.4g}."

    if diverging:
        fill_value = 0.5 * (vmin + vmax)
    else:
        fill_value = vmin
    all_data = _stitch_data_panels(data, fill_value=fill_value)

    fig = plot_imshow(all_data, vmin=vmin, vmax=vmax, cmap=cmap)
    wandb = WandB.get_instance()
    wandb_image = wandb.Image(fig, caption=caption)
    plt.close(fig)

    return wandb_image


def _stitch_data_panels(data: list[list[np.ndarray]], fill_value) -> np.ndarray:
    for row in data:
        if len(row) != len(data[0]):
            raise ValueError("All rows must have the same number of panels.")
    data = [
        [fold_if_healpix_data(arr, fill_value=fill_value) for arr in row]
        for row in data
    ]
    n_rows = len(data)
    n_cols = len(data[0])
    for row in data:
        for arr in row:
            if arr.shape != data[0][0].shape:
                raise ValueError("All panels must have the same shape.")

    stitched_data = np.full(
        (
            n_rows * data[0][0].shape[0] + n_rows - 1,
            n_cols * data[0][0].shape[1] + n_cols - 1,
        ),
        fill_value=fill_value,
    )

    # iterate over rows backwards, as the image starts in the bottom left
    # and moves upwards
    for i, row in enumerate(reversed(data)):
        for j, arr in enumerate(row):
            start_row = i * (arr.shape[0] + 1)
            end_row = start_row + arr.shape[0]
            start_col = j * (arr.shape[1] + 1)
            end_col = start_col + arr.shape[1]
            stitched_data[start_row:end_row, start_col:end_col] = arr

    return stitched_data


def plot_mean_and_samples(
    ax: Any,
    data: xr.DataArray,
    mean_label: str,
    color: str = "cornflowerblue",
    plot_samples: bool = True,
    time_series_dim: str = "year",
    sample_dim: str = "sample",
):
    if time_series_dim in data.dims:
        data.mean(sample_dim).plot(
            ax=ax, x=time_series_dim, label=mean_label, color=color
        )
    if time_series_dim in data.dims and plot_samples:
        for i_sample in range(data.sizes[sample_dim]):
            data.isel(sample=i_sample).plot(
                ax=ax,
                x=time_series_dim,
                color=color,
                alpha=0.4,
                linestyle="-.",
                marker="x",
            )



import os

import pytest
import torch

from fme.core.device import get_device
from fme.core.testing.regression import validate_tensor

from .layers import Context, ContextConfig
from .sfnonet import SphericalFourierNeuralOperatorNet

DIR = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.parametrize(
    "conditional_embed_dim_scalar, conditional_embed_dim_2d, residual_filter_factor",
    [
        (0, 0, 1),
        (16, 0, 1),
        (16, 16, 1),
        (0, 16, 1),
        (16, 0, 4),
    ],
)
def test_can_call_sfnonet(
    conditional_embed_dim_scalar: int,
    conditional_embed_dim_2d: int,
    residual_filter_factor: int,
):
    input_channels = 2
    output_channels = 3
    img_shape = (9, 18)
    n_samples = 4
    device = get_device()
    model = SphericalFourierNeuralOperatorNet(
        params=None,
        embed_dim=16,
        num_layers=2,
        img_shape=img_shape,
        in_chans=input_channels,
        out_chans=output_channels,
        context_config=ContextConfig(
            embed_dim_scalar=conditional_embed_dim_scalar,
            embed_dim_2d=conditional_embed_dim_2d,
        ),
        residual_filter_factor=residual_filter_factor,
    ).to(device)
    x = torch.randn(n_samples, input_channels, *img_shape, device=device)
    context_embedding = torch.randn(
        n_samples, conditional_embed_dim_scalar, device=device
    )
    context_embedding_2d = torch.randn(
        n_samples, conditional_embed_dim_2d, *img_shape, device=device
    )
    context = Context(context_embedding, context_embedding_2d)
    output = model(x, context)
    assert output.shape == (n_samples, output_channels, *img_shape)


def test_sfnonet_output_is_unchanged():
    torch.manual_seed(0)
    input_channels = 2
    output_channels = 3
    img_shape = (9, 18)
    n_samples = 4
    conditional_embed_dim_scalar = 8
    conditional_embed_dim_2d = 16
    device = get_device()
    model = SphericalFourierNeuralOperatorNet(
        params=None,
        embed_dim=16,
        num_layers=2,
        img_shape=img_shape,
        in_chans=input_channels,
        out_chans=output_channels,
        context_config=ContextConfig(
            embed_dim_scalar=conditional_embed_dim_scalar,
            embed_dim_2d=conditional_embed_dim_2d,
        ),
    ).to(device)
    # must initialize on CPU to get the same results on GPU
    x = torch.randn(n_samples, input_channels, *img_shape).to(device)
    context_embedding = torch.randn(n_samples, conditional_embed_dim_scalar).to(device)
    context_embedding_2d = torch.randn(
        n_samples, conditional_embed_dim_2d, *img_shape
    ).to(device)
    context = Context(context_embedding, context_embedding_2d)
    with torch.no_grad():
        output = model(x, context)
    validate_tensor(
        output,
        os.path.join(DIR, "testdata/test_sfnonet_output_is_unchanged.pt"),
    )



import dataclasses
from typing import Literal, Protocol

import torch
import torch.nn.functional as F
import xarray as xr

from fme.core.typing_ import TensorDict


@dataclasses.dataclass
class TimeCoarsenConfig:
    factor: int
    snapshot_names: list[str]
    window_names: list[str]
    window_coarsen_type: Literal["mean"] = "mean"


class Dataset(Protocol):
    """
    Abstract base class for all datasets.

    Meant mainly for internal use in `fme.core.dataset`, to make it clear what methods
    must be implemented to wrap datasets in other dataset types.
    """

    def __len__(self):
        pass

    def __getitem__(self, idx: int) -> tuple[TensorDict, xr.DataArray]:
        pass

    @property
    def sample_start_times(self) -> xr.CFTimeIndex:
        pass

    @property
    def sample_n_times(self) -> int:
        pass


class TimeCoarsenDataset(torch.utils.data.Dataset):
    """
    Wraps a dataset and coarsens each sample in time before passing them on.

    Note that dataset samples do not have a batch dimension.
    """

    def __init__(self, dataset: Dataset, config: TimeCoarsenConfig):
        self.dataset = dataset
        self._config = config

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx: int) -> tuple[TensorDict, xr.DataArray]:
        raw_data, raw_time = self.dataset[idx]
        coarsened_data: dict[str, torch.Tensor] = {}
        n_pooled_timesteps = (
            raw_time.shape[0] // self._config.factor
        ) * self._config.factor
        for name in raw_data.keys():
            n_pooled_timesteps = (
                raw_data[name].shape[0] // self._config.factor
            ) * self._config.factor
            if name in self._config.window_names:
                if self._config.window_coarsen_type == "mean":
                    coarsened_data[name] = F.avg_pool1d(
                        raw_data[name][:n_pooled_timesteps].T,
                        kernel_size=self._config.factor,
                        stride=self._config.factor,
                        ceil_mode=False,
                    ).T
                else:
                    raise ValueError(
                        "Unknown window coarsening type: "
                        f"{self._config.window_coarsen_type}"
                    )
            elif name in self._config.snapshot_names:
                # must take the timestep from the _end_ of a window
                # as windowed data is defined as leading up to the snapshot
                # for that timestep
                coarsened_data[name] = raw_data[name][
                    self._config.factor - 1 : n_pooled_timesteps : self._config.factor
                ]
            else:
                raise ValueError(
                    f"Variable {name} is not a window or snapshot variable, "
                    f"window variables are {self._config.window_names}, "
                    f"snapshot variables are {self._config.snapshot_names}"
                )
        return coarsened_data, raw_time[: n_pooled_timesteps : self._config.factor]

    @property
    def sample_start_times(self) -> xr.CFTimeIndex:
        return self.dataset.sample_start_times

    @property
    def sample_n_times(self) -> int:
        return self.dataset.sample_n_times // self._config.factor



import datetime
from typing import Literal

import pytest
import torch

import fme
from fme.core.coordinates import HybridSigmaPressureCoordinate, LatLonCoordinates
from fme.core.dataset_info import DatasetInfo
from fme.core.normalizer import NetworkAndLossNormalizationConfig, NormalizationConfig
from fme.core.registry import ModuleSelector
from fme.core.typing_ import TensorDict

from .radiation import SeparateRadiationStepConfig

MODULE_TYPES = Literal["AddOne", "SphericalFourierNeuralOperatorNet"]
IMAGE_SHAPE = (4, 2)
MAIN_PROGNOSTIC_NAMES = ["prog_a", "prog_b"]
SHARED_FORCING_NAMES = ["forcing_shared"]
RADIATION_FORCING_NAMES = ["forcing_rad"]
MAIN_DIAGNOSTIC_NAMES = ["diagnostic_main"]
RADIATION_DIAGNOSTIC_NAMES = ["diagnostic_rad"]


class AddOne(torch.nn.Module):
    def forward(self, x):
        return x + 1


def get_tensor_dict(
    names: list[str], img_shape: tuple[int, int], n_samples: int
) -> TensorDict:
    data_dict = {}
    device = fme.get_device()
    for name in names:
        data_dict[name] = torch.rand(
            n_samples,
            *img_shape,
            device=device,
        )
    return data_dict


def get_network_and_loss_normalization_config(
    names: list[str],
    norm_mean: float = 0.0,
) -> NetworkAndLossNormalizationConfig:
    return NetworkAndLossNormalizationConfig(
        network=NormalizationConfig(
            means={name: norm_mean for name in names},
            stds={name: 1.0 for name in names},
        ),
    )


def get_separate_radiation_config(
    module_name: MODULE_TYPES = "SphericalFourierNeuralOperatorNet",
    norm_mean: float = 0.0,
    step_config_kwargs: dict | None = None,
) -> SeparateRadiationStepConfig:
    if step_config_kwargs is None:
        step_config_kwargs = {}

    names = (
        MAIN_PROGNOSTIC_NAMES
        + SHARED_FORCING_NAMES
        + RADIATION_FORCING_NAMES
        + RADIATION_DIAGNOSTIC_NAMES
        + MAIN_DIAGNOSTIC_NAMES
    )
    normalization = get_network_and_loss_normalization_config(
        names=names,
        norm_mean=norm_mean,
    )
    if module_name == "SphericalFourierNeuralOperatorNet":
        config = {
            "scale_factor": 1,
            "embed_dim": 4,
            "num_layers": 2,
        }
        builder = ModuleSelector(type=module_name, config=config)
        radiation_builder = ModuleSelector(type=module_name, config=config)
    elif module_name == "AddOne":
        config = {"module": AddOne()}
        builder = ModuleSelector(type="prebuilt", config=config)
        radiation_builder = ModuleSelector(type="prebuilt", config=config)
    else:
        raise ValueError(f"module_name {module_name!r} is not supported for this test.")

    return SeparateRadiationStepConfig(
        builder=builder,
        radiation_builder=radiation_builder,
        main_prognostic_names=MAIN_PROGNOSTIC_NAMES,
        shared_forcing_names=SHARED_FORCING_NAMES,
        radiation_only_forcing_names=RADIATION_FORCING_NAMES,
        radiation_diagnostic_names=RADIATION_DIAGNOSTIC_NAMES,
        main_diagnostic_names=MAIN_DIAGNOSTIC_NAMES,
        normalization=normalization,
        **step_config_kwargs,
    )


def get_separate_radiation_step(
    module_name: MODULE_TYPES = "SphericalFourierNeuralOperatorNet",
    norm_mean: float = 0.0,
    step_config_kwargs: dict | None = None,
):
    config = get_separate_radiation_config(module_name, norm_mean, step_config_kwargs)
    horizontal_coordinate = LatLonCoordinates(
        lat=torch.zeros(IMAGE_SHAPE[0]), lon=torch.zeros(IMAGE_SHAPE[1])
    )
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.arange(7), bk=torch.arange(7)
    )
    dataset_info = DatasetInfo(
        horizontal_coordinates=horizontal_coordinate,
        vertical_coordinate=vertical_coordinate,
        timestep=datetime.timedelta(hours=6),
    )
    return config.get_step(dataset_info, lambda x: None)


@pytest.mark.parametrize("detach_radiation", [True, False])
def test_detach_radiation(detach_radiation: bool):
    step_config_kwargs = {"detach_radiation": detach_radiation}
    step = get_separate_radiation_step(step_config_kwargs=step_config_kwargs)
    input_data = get_tensor_dict(
        names=step.input_names,
        img_shape=IMAGE_SHAPE,
        n_samples=1,
    )
    input_data["forcing_rad"].requires_grad = True
    output_data = step.step(input_data, input_data)
    for name, value in output_data.items():
        assert value.requires_grad, f"{name} should require grad"
    grad = torch.autograd.grad(
        outputs=output_data["diagnostic_rad"].sum(),
        inputs=input_data["forcing_rad"],
        allow_unused=True,
    )[0]
    assert grad is not None
    # have to call again as torch.autograd.grad frees the graph
    output_data = step.step(input_data, input_data)
    grad = torch.autograd.grad(
        outputs=output_data["diagnostic_main"].sum(),
        inputs=input_data["forcing_rad"],
        allow_unused=True,
    )[0]
    if detach_radiation:
        assert grad is None
    else:
        assert grad is not None


@pytest.mark.parametrize("residual_prediction", [False, True])
def test_residual_prediction(residual_prediction: bool):
    norm_mean = 2.0
    step_config_kwargs = {"residual_prediction": residual_prediction}
    step = get_separate_radiation_step(
        module_name="AddOne",
        norm_mean=norm_mean,
        step_config_kwargs=step_config_kwargs,
    )
    input_data = get_tensor_dict(
        names=step.input_names,
        img_shape=IMAGE_SHAPE,
        n_samples=1,
    )
    output = step.step(input_data, {})

    for name in MAIN_PROGNOSTIC_NAMES:
        if residual_prediction:
            expected_a_output = 2 * input_data[name] + 1 - norm_mean
        else:
            expected_a_output = input_data[name] + 1
        torch.testing.assert_close(output[name], expected_a_output)

    assert not set(SHARED_FORCING_NAMES).intersection(set(output))
    assert not set(RADIATION_FORCING_NAMES).intersection(set(output))



import pytest
import torch

from fme.core.constants import DENSITY_OF_WATER_CM4, SPECIFIC_HEAT_OF_WATER_CM4
from fme.core.coordinates import DepthCoordinate
from fme.core.ocean_data import OceanData


@pytest.mark.parametrize("has_depth_coordinate", [True, False])
def test_column_integrated_ocean_heat_content(has_depth_coordinate: bool):
    """Test column-integrated ocean heat content."""
    n_samples, n_time_steps, nlat, nlon, nlevels = 2, 2, 2, 2, 2
    shape_2d = (n_samples, n_time_steps, nlat, nlon)

    data = {
        "thetao_0": torch.ones(n_samples, n_time_steps, nlat, nlon),
        "thetao_1": torch.ones(n_samples, n_time_steps, nlat, nlon),
    }

    if has_depth_coordinate:
        idepth = torch.tensor([2.5, 10, 20])
        lev_thickness = idepth.diff(dim=-1)
        mask = torch.ones(n_samples, n_time_steps, nlat, nlon, nlevels)
        mask[:, :, 0, 0, 0] = 0.0
        mask[:, :, 0, 0, 1] = 0.0
        mask[:, :, 0, 1, 1] = 0.0

        expected_ohc = torch.tensor(
            SPECIFIC_HEAT_OF_WATER_CM4
            * DENSITY_OF_WATER_CM4
            * n_samples
            * n_time_steps
            * (
                nlat * nlon * lev_thickness.sum()
                - lev_thickness[0]
                - 2 * lev_thickness[1]
            )
        )
        depth_coordinate = DepthCoordinate(idepth, mask)
        ocean_data = OceanData(data, depth_coordinate)
        assert ocean_data.ocean_heat_content.shape == shape_2d
        assert torch.allclose(
            ocean_data.ocean_heat_content.nansum(),
            expected_ohc,
            atol=1e-10,
            equal_nan=True,
        )
    else:
        ocean_data = OceanData(data)
        with pytest.raises(ValueError, match="Depth coordinate must be provided"):
            _ = ocean_data.ocean_heat_content


def test_get_3d_fields():
    """Test getting 3D fields (fields with vertical levels)."""
    n_samples, n_time_steps, nlat, nlon, nlevels = 2, 3, 4, 8, 2
    shape_3d = (n_samples, n_time_steps, nlat, nlon, nlevels)

    data = {
        "thetao_0": torch.rand(n_samples, n_time_steps, nlat, nlon),
        "thetao_1": torch.rand(n_samples, n_time_steps, nlat, nlon),
        "so_0": torch.rand(n_samples, n_time_steps, nlat, nlon),
        "so_1": torch.rand(n_samples, n_time_steps, nlat, nlon),
        "uo_0": torch.rand(n_samples, n_time_steps, nlat, nlon),
        "uo_1": torch.rand(n_samples, n_time_steps, nlat, nlon),
        "vo_0": torch.rand(n_samples, n_time_steps, nlat, nlon),
        "vo_1": torch.rand(n_samples, n_time_steps, nlat, nlon),
    }

    ocean_data = OceanData(data)

    # Test shape of 3D fields
    assert ocean_data.sea_water_potential_temperature.shape == shape_3d
    assert ocean_data.sea_water_salinity.shape == shape_3d
    assert ocean_data.sea_water_x_velocity.shape == shape_3d
    assert ocean_data.sea_water_y_velocity.shape == shape_3d


def test_get_2d_fields():
    """Test getting 2D surface fields."""
    n_samples, n_time_steps, nlat, nlon = 2, 3, 4, 8
    shape_2d = (n_samples, n_time_steps, nlat, nlon)

    data = {
        "sst": torch.rand(n_samples, n_time_steps, nlat, nlon),
        "zos": torch.rand(n_samples, n_time_steps, nlat, nlon),
    }

    ocean_data = OceanData(data)

    # Test shape of 2D fields
    assert ocean_data.sea_surface_temperature.shape == shape_2d
    assert ocean_data.sea_surface_height_above_geoid.shape == shape_2d


def test_missing_field():
    """Test that accessing a missing field raises KeyError."""
    data = {"sst": torch.rand(2, 3, 4, 8)}
    ocean_data = OceanData(data)

    with pytest.raises(KeyError, match="thetao_"):
        _ = ocean_data.sea_water_potential_temperature


@pytest.mark.parametrize("missing_layer", [True, False])
def test_keyerror_when_missing_3d_layer(missing_layer: bool):
    """Test that missing a layer in a 3D field raises ValueError."""
    n_samples, n_time_steps, nlat, nlon = 2, 3, 4, 8

    def _get_data(missing_layer: bool):
        data = {
            "thetao_1": torch.rand(n_samples, n_time_steps, nlat, nlon),
        }
        if not missing_layer:
            data["thetao_0"] = torch.rand(n_samples, n_time_steps, nlat, nlon)
        return data

    ocean_data = OceanData(_get_data(missing_layer))

    if not missing_layer:
        assert ocean_data.sea_water_potential_temperature.shape == (
            n_samples,
            n_time_steps,
            nlat,
            nlon,
            2,
        )
    else:
        with pytest.raises(ValueError, match="Missing level 0 in thetao_ levels"):
            _ = ocean_data.sea_water_potential_temperature


def test_getitem():
    """Test the __getitem__ method."""
    data = {"sst": torch.rand(2, 3, 4, 8)}
    ocean_data = OceanData(data)

    assert torch.equal(
        ocean_data["sea_surface_temperature"], ocean_data.sea_surface_temperature
    )

    with pytest.raises(
        AttributeError, match="object has no attribute 'nonexistent_field'"
    ):
        _ = ocean_data["nonexistent_field"]



# This script is used to compute a training dataset from the "raw"
# E3SMv2 data stored in netCDF form on LC lustre.

# The dependencies of this script are installed in the "fv3net" conda environment
# which can be installed using fv3net's Makefile. See
# https://github.com/ai2cm/fv3net/blob/8ed295cf0b8ca49e24ae5d6dd00f57e8b30169ac/Makefile#L310

# On the Perlmutter cpu partition using --n-workers=32, this script takes about
# 1hr 10min to complete, using a max of about 100 GB of memory for 21 years of
# input data.

import os
import time
from glob import glob
from itertools import chain
from typing import Callable, List, MutableMapping, Optional, Sequence, Tuple

import click
import numpy as np
import xarray as xr
import xpartition  # noqa: F401
from compute_dataset import (
    DatasetComputationConfig,
    DatasetConfig,
    assert_column_integral_of_moisture_is_conserved,
    assert_global_dry_air_mass_conservation,
    assert_global_moisture_conservation,
    compute_column_advective_moisture_tendency,
    compute_column_moisture_integral,
    compute_specific_total_water,
    compute_tendencies,
    compute_vertical_coarsening,
)
from dask.diagnostics import ProgressBar
from dask.distributed import Client
from xtorch_harmonics import roundtrip_filter

# default paths for input/output; can be changed when calling this script
INPUT_DIR = "/global/cfs/cdirs/e3sm/golaz/E3SM/fme/20230614.v2.LR.F2010/post/atm/180x360_gaussian/ts"  # noqa: E501
OUTPUT_URL = "/pscratch/sd/j/jpduncan/ai2/zarr/2023-11-22-e3smv2-vertically-resolved-1deg-fme-dataset.zarr"  # noqa: E501

REFERENCE_PRESSURE = 1e5  # Pa
LIQUID_PRECIP_DENSITY = 1e3  # kg/m**3
LATENT_HEAT_OF_VAPORIZATION = 2.501e6  # J/kg

# derived variable names
SURFACE_UP_LONGWAVE_FLUX = "surface_upward_longwave_flux"
SURFACE_UP_SHORTWAVE_FLUX = "surface_upward_shortwave_flux"
TOA_UP_SHORTWAVE_FLUX = "top_of_atmos_upward_shortwave_flux"

RAD_FLUX_FORMULAS = {
    SURFACE_UP_LONGWAVE_FLUX: (lambda x, y: x + y, "FLNS", "FLDS"),
    SURFACE_UP_SHORTWAVE_FLUX: (lambda x, y: x - y, "FSDS", "FSNS"),
    TOA_UP_SHORTWAVE_FLUX: (lambda x, y: x - y, "SOLIN", "FSNTOA"),
}

SURFACE_PRECIPITATION = "surface_precipitation_rate"
PRECIPITABLE_WATER_PATH_NAMES = ["TMQ", "TGCLDLWP", "TGCLDIWP"]

DROP_VARIABLE_NAMES = {
    "2D": [  # variables to drop when opening 2D vars
        "time_bnds",
        "lat_bnds",
        "lon_bnds",
        "area",
        "gw",
        "elevation",
    ],
    "3D": [  # variables to drop when opening 3D vars
        "P0",
        "PS",
        "time_bnds",
        "lat_bnds",
        "lon_bnds",
        "area",
        "gw",
        "hyam",
        "hybm",
    ],
}


def expand_names_by_level(variables: MutableMapping[str, List[str]]) -> List[str]:
    names = []
    for var_name, levels in variables.items():
        names.extend([f"{var_name}{lev}" for lev in levels])
    return names


def get_nc_paths(
    base_dir: str, var_names: Sequence[str]
) -> MutableMapping[str, List[str]]:
    paths = {
        var_name: sorted(list(glob(os.path.join(base_dir, f"{var_name}_*.nc"))))
        for var_name in var_names
    }
    return paths


def get_time_invariant_nc_paths(
    base_dir: Optional[str],
) -> MutableMapping[str, List[str]]:
    paths = {
        "time_invariant": list(glob(os.path.join(base_dir, f"*.nc")))  # type: ignore
    }
    return paths


def open_dataset(
    dataset_dirs: MutableMapping[str, str],
    config: DatasetComputationConfig,
) -> xr.Dataset:
    """Open datasets from NetCDF files in directory that match input variable names."""
    var_paths: MutableMapping[str, List[str]] = {}
    input_variable_names = config.variable_sources
    for key in dataset_dirs.keys():
        var_paths.update(get_nc_paths(dataset_dirs[key], input_variable_names[key]))
    var_paths.update(get_time_invariant_nc_paths(config.time_invariant_dir))
    print(
        f"Opening {len(list(chain.from_iterable(var_paths.values())))} files with "
        f"{len(var_paths.keys())} vars..."
    )
    standard_names = config.standard_names
    varnames_3D = [
        standard_names.air_temperature,
        standard_names.eastward_wind,
        standard_names.northward_wind,
    ] + standard_names.water_species
    chunks = config.chunking.get_chunks(config.standard_names)
    datasets = {}
    start = time.time()
    if "time_invariant" in var_paths:
        for path in var_paths["time_invariant"]:
            ds = xr.open_dataset(path, decode_timedelta=False).drop(
                DROP_VARIABLE_NAMES["2D"], errors="ignore"
            )
            if "time" in ds.coords:
                ds = ds.isel(time=0, drop=True)
            for varname in input_variable_names["time_invariant"]:
                if varname not in datasets and varname in ds.variables:
                    datasets[varname] = ds
        del var_paths["time_invariant"]
    for varname, paths in var_paths.items():
        var_start = time.time()
        if varname in varnames_3D:
            drop_vars = DROP_VARIABLE_NAMES["3D"]
        else:
            drop_vars = DROP_VARIABLE_NAMES["2D"]
        datasets[varname] = xr.open_mfdataset(
            paths,
            chunks=chunks,
            data_vars="minimal",
            coords="minimal",
            parallel=True,
        ).drop(drop_vars, errors="ignore")
        print(f"{varname} files opened in {time.time() - var_start:.2f} s...")
    print(f"All files opened in {time.time() - start:.2f} s. Merging...")
    return xr.merge(datasets.values(), compat="override", join="override")


def compute_pressure_thickness(
    ds: xr.Dataset,
    vertical_dim: str,
    vertical_interface_dim: str,
    hybrid_level_coeffs: List[str],
    reference_pressure: float,
    surface_pressure: str,
    output_name: str,
):
    hyai, hybi = hybrid_level_coeffs
    sfc_pressure = ds[surface_pressure].expand_dims(
        {vertical_interface_dim: ds[vertical_interface_dim]}, axis=3
    )
    phalf = sfc_pressure * ds[hybi] + reference_pressure * ds[hyai]
    thickness = (
        phalf.diff(dim=vertical_interface_dim)
        .rename({vertical_interface_dim: vertical_dim})
        .rename(output_name)
        .assign_coords({vertical_dim: (vertical_dim, ds[vertical_dim].values)})
    )
    thickness.attrs["units"] = "Pa"
    thickness.attrs["long_name"] = output_name.replace("_", " ")
    return ds.assign({output_name: thickness})


def compute_coarse_ak_bk(
    ds: xr.Dataset,
    interface_indices: Sequence[Tuple[int, int]],
    z_dim: str,
    hybrid_level_coeffs: List[str],
    reference_pressure: float = REFERENCE_PRESSURE,
):
    """Return dataset with scalar ak and bk coordinates that define coarse
    interfaces, which should match the conventions used in compute_dataset_fv3gfs.py.

    Args:
        ds: xr.Dataset with hybrid level coefficients named like hybrid_level_coeffs.
        interface_indices: list of tuples of indices of the interfaces in the vertical.
        z_dim: name of dimension along which ak and bk are defined.
        hybrid_level_coeffs: Hybrid level coeff names in ds, ordered like ["ak", "bk"].

    Returns:
        xr.Dataset with ak and bk variables as scalars labeled as ak_0, bk_0, etc.

    Note:
        The ak and bk variables will have one more vertical level than the other 3D
        variables since they represent the interfaces between levels.

    """
    data = {}
    hyai, hybi = hybrid_level_coeffs
    for i, (start, end) in enumerate(interface_indices):
        data[f"ak_{i}"] = ds[hyai].isel({z_dim: start}) * reference_pressure
        data[f"bk_{i}"] = ds[hybi].isel({z_dim: start})
        if i == len(interface_indices) - 1:
            data[f"ak_{i + 1}"] = ds[hyai].isel({z_dim: end}) * reference_pressure
            data[f"bk_{i + 1}"] = ds[hybi].isel({z_dim: end})
    for i in range(len(interface_indices) + 1):
        data[f"ak_{i}"].attrs["units"] = "Pa"
        data[f"bk_{i}"].attrs["units"] = ""  # unitless quantity
        for name in ["ak", "bk"]:
            data[f"{name}_{i}"] = data[f"{name}_{i}"].drop(z_dim)
    return xr.Dataset(data).compute()


def compute_surface_precipitation_rate(
    ds: xr.Dataset,
    total_precip_rate_name,
    liquid_precip_density: float = LIQUID_PRECIP_DENSITY,
    output_name: str = SURFACE_PRECIPITATION,
):
    precip_mass_flux = ds[total_precip_rate_name] * liquid_precip_density
    precip_mass_flux.attrs["units"] = "kg/m2/s"
    precip_mass_flux.attrs["long_name"] = output_name.replace("_", " ")
    return ds.assign({output_name: precip_mass_flux})


def compute_precipitable_water_path(
    ds: xr.Dataset,
    output_name: str,
    precipitable_water_path_names: List[str],
):
    water_path = sum([ds[name] for name in precipitable_water_path_names])
    water_path.attrs["units"] = "kg/m2"
    water_path.attrs["long_name"] = output_name.replace("_", " ")
    return ds.assign({output_name: water_path})


def compute_rad_fluxes(
    ds: xr.Dataset,
    rad_flux_formulas: MutableMapping[
        str, Tuple[Callable, str, str]
    ] = RAD_FLUX_FORMULAS,
) -> xr.Dataset:
    fluxes = {}
    for output_name, formula in rad_flux_formulas.items():
        fluxes[output_name] = formula[0](ds[formula[1]], ds[formula[2]])
        fluxes[output_name].attrs["long_name"] = output_name.replace("_", " ")
        fluxes[output_name].attrs["units"] = ds[formula[1]].attrs["units"]
    return ds.assign(fluxes)


def construct_lazy_dataset(
    config: DatasetComputationConfig,
    dataset_dirs: MutableMapping[str, str],
) -> xr.Dataset:
    start = time.time()
    standard_names = config.standard_names
    print(f"Opening dataset...")
    ds = open_dataset(dataset_dirs, config)
    print(f"Dataset opened in {time.time() - start:.2f} s total.")
    print(f"Input dataset size is {ds.nbytes / 1e9} GB")
    if config.roundtrip_fraction_kept is not None:
        ds = roundtrip_filter(
            ds,
            lat_dim=standard_names.latitude_dim,
            lon_dim=standard_names.longitude_dim,
            fraction_modes_kept=config.roundtrip_fraction_kept,
        )

    ds = compute_pressure_thickness(
        ds,
        vertical_dim=standard_names.vertical_dim,
        vertical_interface_dim=standard_names.vertical_interface_dim,
        hybrid_level_coeffs=standard_names.hybrid_level_coeffs,
        reference_pressure=REFERENCE_PRESSURE,
        surface_pressure=standard_names.surface_pressure,
        output_name=standard_names.pressure_thickness,
    )
    ds = compute_rad_fluxes(ds)
    ds = compute_surface_precipitation_rate(
        ds,
        total_precip_rate_name=standard_names.precip_rate,
    )
    water_species_name = [
        item for item in standard_names.water_species if item.lower() != "none"
    ]
    ds = compute_specific_total_water(
        ds,
        water_condensate_names=water_species_name,
        output_name=standard_names.specific_total_water,
    )
    ds = compute_vertical_coarsening(
        ds,
        vertically_resolved_names=standard_names.vertically_resolved,
        interface_indices=config.vertical_coarsening_indices,
        dim=standard_names.vertical_dim,
        pressure_thickness_name=standard_names.pressure_thickness,
        validate_indices=config.validate_vertical_coarsening_indices,
    )
    ds = compute_column_moisture_integral(
        ds,
        input_name=standard_names.specific_total_water,
        output_name=standard_names.total_water_path,
        pressure_thickness_name=standard_names.pressure_thickness,
        dim=standard_names.vertical_dim,
    )
    ds = compute_tendencies(
        ds,
        time_derivative_names=standard_names.time_derivative_names,
        dim=standard_names.time_dim,
    )
    ds = compute_column_advective_moisture_tendency(
        ds,
        pwat_tendency=standard_names.pwat_tendency,
        latent_heat_flux=standard_names.latent_heat_flux,
        precip=standard_names.precip_rate,
        latent_heat_of_vaporization=LATENT_HEAT_OF_VAPORIZATION,
    )
    ak_bk_ds = compute_coarse_ak_bk(
        ds,
        interface_indices=config.vertical_coarsening_indices,
        z_dim=standard_names.vertical_interface_dim,
        hybrid_level_coeffs=standard_names.hybrid_level_coeffs,
    )
    ds = xr.merge([ds, ak_bk_ds])
    ds_dirs = list(dataset_dirs.values())
    chunks = config.chunking.get_chunks(config.standard_names)
    ds = ds.chunk(chunks).astype(np.float32)
    ds.attrs["history"] = (
        "Dataset computed by full-model/scripts/e3smv2_data_process"
        "/compute_dataset_e3smv2.py"
        f" script, using inputs from the following directories: {ds_dirs}."
    )
    ds.attrs["vertical_coordinate"] = (
        "The pressure at level interfaces can by computed as "
        "p_i = ak_i + bk_i * PS, where PS is the surface pressure and the "
        "p_i pressure corresponds to the interface at the top of the i'th finite "
        "volume layer, counting down from the top of atmosphere."
    )
    return ds


@click.command()
@click.option("--config", help="Path to dataset configuration YAML file.")
@click.option(
    "-i",
    "--input-dir",
    default=INPUT_DIR,
    help="Directory in which to find time-varying input ncs.",
)
@click.option("-o", "--output", default=OUTPUT_URL, help="URL to write output to.")
@click.option("--debug", is_flag=True, help="Print metadata instead of writing output.")
@click.option("--subsample", is_flag=True, help="Subsample the data before writing.")
@click.option("--check-conservation", is_flag=True, help="Check conservation.")
@click.option(
    "--water-budget-dataset",
    is_flag=True,
    help="Create a dataset of 2D vars for checking the water budget.",
)
@click.option("--n-workers", default=4, help="Number of Dask workers.")
def main(
    config,
    input_dir,
    output,
    debug,
    subsample,
    check_conservation,
    water_budget_dataset,
    n_workers,
):
    xr.set_options(keep_attrs=True)
    _ = Client(n_workers=n_workers)
    config = DatasetConfig.from_file(config).dataset_computation
    standard_names = config.standard_names
    dataset_dirs = {}
    for key in config.variable_sources.keys():
        if key != "time_invariant":
            dataset_dirs[key] = os.path.join(input_dir, key)
    ds = construct_lazy_dataset(config, dataset_dirs)
    if subsample:
        ds = ds.isel(time=slice(10, 13))
    if check_conservation:
        # these currently fail
        ds = compute_precipitable_water_path(
            ds,
            output_name=standard_names.precipitable_water_path,
            precipitable_water_path_names=PRECIPITABLE_WATER_PATH_NAMES,
        )
        assert_column_integral_of_moisture_is_conserved(
            ds, standard_names.precipitable_water_path, standard_names.total_water_path
        )
        assert_global_dry_air_mass_conservation(
            ds,
            dims=standard_names.horizontal_dims,
            surface_pressure_name=standard_names.surface_pressure,
            total_water_path_name=standard_names.total_water_path,
            latitude_dim=standard_names.latitude_dim,
            time_dim=standard_names.time_dim,
        )
        assert_global_moisture_conservation(
            ds,
            dims=standard_names.horizontal_dims,
            latitude_dim=standard_names.latitude_dim,
            total_water_path_name=standard_names.total_water_path,
            latent_heat_flux_name=standard_names.latent_heat_flux,
            latent_heat_of_vaporization=LATENT_HEAT_OF_VAPORIZATION,
            precip_rate_name=standard_names.precip_rate,
            time_dim=standard_names.time_dim,
        )
    if water_budget_dataset:
        water_budget_dataset_vars = [
            standard_names.surface_pressure,
            standard_names.precipitable_water_path,
            standard_names.precip_rate,
            standard_names.latent_heat_flux,
            "PRECSC",
            "PRECSL",
            "QFLX",
            "TMQ",
            "TGCLDLWP",
            "TGCLDIWP",
        ]

        ds = ds[water_budget_dataset_vars]
    else:
        dropped_variables = (
            [
                item
                for item in standard_names.dropped_variables
                if item.lower() != "none"
            ]
            + standard_names.hybrid_level_coeffs
            + [
                standard_names.precip_rate,
                standard_names.vertical_interface_dim,
                "TMQ",
                "TGCLDLWP",
                "TGCLDIWP",
                "FLNS",
                "FSNS",
                "FSNTOA",
                "PRECSC",
                "PRECSL",
                "QFLX",
            ]
        )
        ds = ds.drop(dropped_variables, errors="ignore")
    print(f"Output dataset size is {ds.nbytes / 1e9} GB")
    if debug:
        with xr.set_options(display_max_rows=500):
            print(ds)
    else:
        ds.partition.initialize_store(output)
        for i in range(config.n_split):
            print(f"Writing segment {i + 1} / {config.n_split}")
            with ProgressBar():
                ds.partition.write(
                    output, config.n_split, ["time"], i, collect_variable_writes=True
                )


if __name__ == "__main__":
    main()



import argparse
import dataclasses
from typing import Callable, List, Tuple

import dacite
import fsspec
import numpy as np
import xarray as xr
import yaml
from scipy.optimize import curve_fit

from fme.core.wandb import WandB


@dataclasses.dataclass
class DataConfig:
    stats_path: str
    names: List[str]

    def get_datasets(self) -> Tuple[xr.Dataset, xr.Dataset]:
        if self.stats_path.endswith("/"):
            stats_path = self.stats_path
        else:
            stats_path = self.stats_path + "/"
        return (
            xr.open_dataset(fsspec.open(stats_path + "rmse_means.nc").open()),
            xr.open_dataset(fsspec.open(stats_path + "rmse_stdevs.nc").open()),
        )


def fit(x: np.ndarray, y: np.ndarray) -> Callable[[np.ndarray], np.ndarray]:
    """
    Fits given data to a function of the form y = c / sqrt(x) + b.
    """

    def model(x, c, b):
        return c * 1 / np.sqrt(x) + b

    popt, _ = curve_fit(model, x, y)

    c, b = popt

    def fitted_model(x):
        return model(x, c, b)

    return fitted_model


def main(
    rmse_means: xr.Dataset,
    rmse_stdevs: xr.Dataset,
    varnames: list[str],
    window_years: int,
    batches: int,
    project: str,
    entity: str,
    name: str,
):
    rmse_means = rmse_means.load()
    rmse_stdevs = rmse_stdevs.load()
    noise_floors = {}
    stds = {}
    for varname in varnames:
        x = rmse_means[varname].window_size
        y = rmse_means[varname].values
        fitted_model = fit(x, y)
        noise_floors[varname] = fitted_model(window_years)
        x = rmse_stdevs[varname].window_size
        y = rmse_stdevs[varname].values
        fitted_model = fit(x, y)
        stds[varname] = fitted_model(window_years)

    wandb = WandB.get_instance()
    wandb.configure(log_to_wandb=True)
    config = {
        "window_years": window_years,
        "batches": batches,
    }

    wandb.init(
        name=name + "-plus-std",
        job_type="training",
        config=config,
        project=project,
        entity=entity,
        reinit=True,
    )
    for i in [0, batches]:
        for varname in noise_floors:
            wandb.log(
                {
                    f"inference/time_mean/rmse/{varname}": noise_floors[varname]
                    + stds[varname],
                },
                step=i,
                sleep=0.1,
            )

    wandb.init(
        name=name,
        job_type="training",
        config=config,
        project=project,
        entity=entity,
        reinit=True,
    )
    for i in [0, batches]:
        for varname in noise_floors:
            wandb.log(
                {
                    f"inference/time_mean/rmse/{varname}": noise_floors[varname],
                },
                step=i,
                sleep=0.1,
            )

    wandb.init(
        name=name + "-minus-std",
        job_type="training",
        config=config,
        project=project,
        entity=entity,
        reinit=True,
    )
    for i in [0, batches]:
        for varname in noise_floors:
            wandb.log(
                {
                    f"inference/time_mean/rmse/{varname}": noise_floors[varname]
                    - stds[varname],
                },
                step=i,
                sleep=0.1,
            )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("data_config", type=str)
    parser.add_argument("name", type=str)
    parser.add_argument("window_years", type=int)
    parser.add_argument("batches", type=int)
    parser.add_argument("--project", type=str, default="fourcastnet")
    parser.add_argument("--entity", type=str, default="ai2cm")
    args = parser.parse_args()

    with open(args.data_config, "r") as f:
        data_config = dacite.from_dict(
            DataConfig,
            yaml.safe_load(f),
        )

    rmse_means, rmse_stdevs = data_config.get_datasets()
    main(
        rmse_means=rmse_means,
        rmse_stdevs=rmse_stdevs,
        varnames=data_config.names,
        window_years=args.window_years,
        batches=args.batches,
        project=args.project,
        entity=args.entity,
        name=args.name,
    )



import torch
import xarray as xr

from fme.core.histogram import ComparedDynamicHistograms
from fme.core.typing_ import TensorMapping


class HistogramAggregator:
    def __init__(self):
        self._histograms = ComparedDynamicHistograms(n_bins=200, percentiles=[99.9999])

    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
        i_time_start: int = 0,
    ):
        self._histograms.record_batch(target_data, gen_data)

    @torch.no_grad()
    def get_logs(self, label: str):
        logs = self._histograms.get_wandb()
        if label != "":
            logs = {f"{label}/{k}": v for k, v in logs.items()}
        return logs

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset:
        return self._histograms.get_dataset()



import numpy as np
import pytest
import torch
import torch_harmonics

import fme
from fme.core.constants import GRAVITY
from fme.core.device import get_device
from fme.core.metrics import (
    net_surface_energy_flux,
    quantile,
    spherical_power_spectrum,
    surface_pressure_due_to_dry_air,
)


def _get_lats(num_lat: int):
    """A simple example of latitudes."""
    return torch.linspace(-89.5, 89.5, num_lat)


test_cases = (
    "variable, time, lats, n_lon",
    [
        (2, 1, _get_lats(2), 1),
        (1, 2, _get_lats(2), 1),
        (1, 2, _get_lats(2), 4),
    ],
)


@pytest.mark.parametrize(
    "num_lat, num_lon, expected",
    [
        (2, 1, torch.tensor([[0.5], [0.5]])),
        (2, 2, torch.tensor([[0.25, 0.25], [0.25, 0.25]])),
        (
            2,
            4,
            torch.tensor(
                [[0.1250, 0.1250, 0.1250, 0.1250], [0.1250, 0.1250, 0.1250, 0.1250]]
            ),
        ),
    ],
)
def test_spherical_area_weights(num_lat, num_lon, expected):
    """Tests the shapes and a couple simple cases of the spherical area weights."""
    lats = _get_lats(num_lat)
    result = fme.spherical_area_weights(lats, num_lon)
    assert torch.all(torch.isclose(result, expected))


def test_spherical_area_weights_batch_dim():
    """Tests the spherical area weights with a batch dimension."""
    num_lat, num_lon = 2, 4
    batch_size = 3
    lats = _get_lats(num_lat).unsqueeze(0).expand(batch_size, num_lat)
    result = fme.spherical_area_weights(lats, num_lon)
    expected = (
        fme.spherical_area_weights(lats[0], num_lon)
        .unsqueeze(0)
        .expand(batch_size, -1, -1)
    )
    assert result.shape == (batch_size, num_lat, num_lon)
    assert torch.equal(result, expected)


@pytest.mark.parametrize(*test_cases)
def test_weighted_nanmean(variable, time, lats, n_lon):
    """Tests the weighted nanmean over test_cases, with NaNs in the input
    tensor. See also test_weighted_nanmean_ignores_nans below."""
    x = torch.randn(time, variable, len(lats), n_lon)
    x_nan = x.clone()
    x_nan[0, 0, 0, 0] = float("nan")

    mask = ~torch.isnan(x_nan)
    expected_result = (x * mask).sum(dim=(0, 2, 3)) / mask.sum(dim=(0, 2, 3))

    result = fme.weighted_nanmean(x_nan, dim=(0, 2, 3))
    assert torch.allclose(
        result, expected_result
    ), "unweighted mean with NaNs should ignore NaNs"

    # weighting with ones shouldn't change the result
    weights = torch.ones(len(lats), n_lon)
    result = fme.weighted_nanmean(x_nan, weights, dim=(0, 2, 3))
    assert torch.allclose(
        result, expected_result
    ), "weighted mean with NaNs weights by ones should be the same as unweighted"

    weights = torch.randn(len(lats), n_lon)
    weights_like_nan = weights.expand(x.shape).clone()
    weights_like_nan[0, 0, 0, 0] = 0.0

    expected_result = fme.weighted_nanmean(x, weights_like_nan, dim=(0, 2, 3))
    result = fme.weighted_nanmean(x_nan, weights, dim=(0, 2, 3))
    assert torch.allclose(
        result, expected_result
    ), "weighted mean of NaNs should act like zero weight in masked regions"


@pytest.mark.parametrize(
    "x_nan, weights, expected_result",
    [
        pytest.param(
            torch.tensor([float("nan"), 0, 1]),
            None,
            torch.tensor(1 / 2),
            id="unweighted nanmean ignores nans",
        ),
        pytest.param(
            torch.tensor([float("nan"), 0, 1]),
            torch.tensor([1, 1, 1]),
            torch.tensor(1 / 2),
            id="weighted nanmean still ignores nans, shape: (3,)",
        ),
        pytest.param(
            torch.tensor([[float("nan"), 1, 2], [float("nan"), 3, 4]]),
            torch.tensor([0, 1, 2]),
            torch.tensor([(1 + 4) / 3, (3 + 8) / 3]),
            id="zero weight nan ignored, shape: (2, 3)",
        ),
        pytest.param(
            torch.tensor([[float("nan"), 1, 2], [2, 3, 4]]),
            torch.tensor([1, 1, 2]),
            torch.tensor([(1 + 4) / 3, (2 + 3 + 8) / 4]),
            id="nonzero weighted nan ignored, shape: (2, 3)",
        ),
    ],
)
def test_weighted_nanmean_ignores_nans(x_nan, weights, expected_result):
    """Tests the weighted nanmean for a few simple test cases, with NaNs in the
    input tensor.

    weighted_nanmean always ignores NaNs, regardless of their assigned weight.
    """
    result = fme.weighted_nanmean(x_nan, weights=weights, dim=-1)
    torch.testing.assert_close(
        result,
        expected_result,
    )


@pytest.mark.parametrize(
    "x_nan, weights, expected_result",
    [
        pytest.param(
            torch.tensor([float("nan"), 0, 1]),
            None,
            torch.tensor(float("nan")),
            id="unweighted sum with nan is nan",
        ),
        pytest.param(
            torch.tensor([float("nan"), 0, 1]),
            torch.tensor([0, 1, 1]),
            torch.tensor(1.0),
            id="zero weight nan ignored, shape: (3,)",
        ),
        pytest.param(
            torch.tensor([[float("nan"), 1, 2], [float("nan"), 3, 4]]),
            torch.tensor([0, 1, 2]),
            torch.tensor([float(1 + 4), float(3 + 8)]),
            id="zero weight nan ignored, shape: (2, 3)",
        ),
        pytest.param(
            torch.tensor([[float("nan"), 1, 2], [2, 3, 4]]),
            torch.tensor([1, 1, 2]),
            torch.tensor([float("nan"), float(2 + 3 + 8)]),
            id="nonzero weighted nan survives, shape: (2, 3)",
        ),
    ],
)
def test_weighted_sum_with_nans(x_nan, weights, expected_result):
    """Tests the weighted sum for a few simple test cases, with NaNs in the
    input tensor.

    weighted_sum does not ignore NaNs unless they have zero weight.
    """
    result = fme.weighted_sum(x_nan, weights=weights, dim=-1)
    torch.testing.assert_close(
        result,
        expected_result,
        equal_nan=True,
    )


@pytest.mark.parametrize(
    "x_nan, weights, expected_result",
    [
        pytest.param(
            torch.tensor([float("nan"), 0, 1]),
            None,
            torch.tensor(float("nan")),
            id="unweighted mean with nan is nan",
        ),
        pytest.param(
            torch.tensor([float("nan"), 0, 1]),
            torch.tensor([0, 1, 1]),
            torch.tensor(1 / 2),
            id="zero weight nan ignored, shape: (3,)",
        ),
        pytest.param(
            torch.tensor([[float("nan"), 1, 2], [float("nan"), 3, 4]]),
            torch.tensor([0, 1, 2]),
            torch.tensor([(1 + 4) / 3, (3 + 8) / 3]),
            id="zero weight nan ignored, shape: (2, 3)",
        ),
        pytest.param(
            torch.tensor([[float("nan"), 1, 2], [2, 3, 4]]),
            torch.tensor([1, 1, 2]),
            torch.tensor([float("nan"), (2 + 3 + 8) / 4]),
            id="nonzero weighted nan survives, shape: (2, 3)",
        ),
    ],
)
def test_weighted_mean_with_nans(x_nan, weights, expected_result):
    """Tests the weighted mean for a few simple test cases, with NaNs in the
    input tensor.

    weighted_mean does not ignore NaNs unless they have zero weight.
    """
    result = fme.weighted_mean(x_nan, weights=weights, dim=-1)
    torch.testing.assert_close(
        result,
        expected_result,
        equal_nan=True,
    )


def test_weighted_std_constant_weights():
    """Tests the weighted std with constant weights."""
    torch.manual_seed(0)
    n_lat = 4
    n_lon = 5
    x = torch.randn(2, 3, n_lat, n_lon)
    weights = torch.ones(n_lat, n_lon)
    result = fme.weighted_std(x, weights, dim=(0, 2, 3))
    # must use numpy std here as we want the standard deviation of
    # the sample, not the population
    unweighted_std = np.std(x.cpu().numpy(), axis=(0, 2, 3))
    assert torch.all(
        torch.isclose(result, torch.as_tensor(unweighted_std))
    ), "Weighted std with constant weights should be same as unweighted std."


def test_weighted_std_none_weights():
    """Tests the weighted std without passing weights."""
    torch.manual_seed(0)
    n_lat = 4
    n_lon = 5
    x = torch.randn(2, 3, n_lat, n_lon)
    weights = None
    result = fme.weighted_std(x, weights, dim=(0, 2, 3))
    # must use numpy std here as we want the standard deviation of
    # the sample, not the population
    unweighted_std = np.std(x.cpu().numpy(), axis=(0, 2, 3))
    assert torch.all(
        torch.isclose(result, torch.as_tensor(unweighted_std))
    ), "Weighted std with constant weights should be same as unweighted std."


@pytest.mark.parametrize(*test_cases)
def test_weighted_mean_bias(variable, time, lats, n_lon):
    """Tests the weighted mean bias for a few simple test cases."""
    x = torch.randn(time, variable, len(lats), n_lon)
    y = torch.randn(time, variable, len(lats), n_lon)
    weights = fme.spherical_area_weights(lats, n_lon)

    result = fme.weighted_mean_bias(x, x.clone(), weights, dim=(0, 2, 3))
    assert torch.all(
        torch.isclose(result, torch.tensor(0.0))
    ), "Weighted global mean bias between identical tensors should be zero."
    assert result.shape == (variable,), "You should be able to specify time as dim = 1."

    x = torch.zeros(time, variable, len(lats), n_lon)
    y = torch.ones(time, variable, len(lats), n_lon)

    result = fme.weighted_mean_bias(x, y, weights)
    assert torch.all(
        torch.isclose(result, torch.Tensor([1.0]))
    ), "The weighted mean of a constant should be that constant."

    result = fme.weighted_mean_bias(x, y)
    assert result.shape == tuple(), "Should also work if you do not specify weights."

    x = torch.randn(variable, time, n_lon, len(lats))
    y = torch.randn(variable, time, n_lon, len(lats))
    result = fme.weighted_mean_bias(x, x.clone(), weights.t(), dim=(1, 2, 3))
    assert torch.all(
        torch.isclose(result, torch.tensor(0.0))
    ), "Weighted global mean bias between identical tensors should be zero."
    assert result.shape == (
        variable,
    ), "Swapping dims shouldn't change the final shape."


@pytest.mark.parametrize(*test_cases)
def test_root_mean_squared_error(variable, time, lats, n_lon):
    """Tests the mean squared error for a few simple test cases."""
    x = torch.randn(variable, time, len(lats), n_lon)
    random_weights = torch.rand(len(lats), n_lon)

    result = fme.root_mean_squared_error(x, x.clone(), dim=(0, 2, 3))
    assert torch.all(
        torch.isclose(result, torch.tensor(0.0))
    ), "Root mean squared error between identical tensors should be zero."

    result = fme.root_mean_squared_error(
        torch.zeros(variable, time, len(lats), n_lon),
        torch.ones(variable, time, len(lats), n_lon),
    )
    assert torch.all(
        torch.isclose(result, torch.tensor(1.0))
    ), "Root mean squared error between zero and one should be one."

    result = fme.root_mean_squared_error(
        torch.zeros(variable, time, len(lats), n_lon),
        torch.ones(variable, time, len(lats), n_lon),
        weights=random_weights,
    )
    assert torch.isclose(
        result, torch.tensor([1.0]).sqrt()
    ), "Root mean squared error between zero and one should be one."


@pytest.mark.parametrize(*test_cases)
def test_rmse_of_time_mean(variable, time, lats, n_lon):
    x = torch.randn(variable, time, len(lats), n_lon)
    random_weights = torch.rand(len(lats), n_lon)

    result = fme.rmse_of_time_mean(x, x.clone(), time_dim=1)
    torch.testing.assert_close(
        result,
        torch.zeros((variable,)),
        msg="RMSE of time mean between identical tensors should be zero.",
    )

    result = fme.rmse_of_time_mean(
        torch.zeros(variable, time, len(lats), n_lon),
        torch.ones(variable, time, len(lats), n_lon),
        weights=random_weights,
        time_dim=1,
    )
    torch.testing.assert_close(
        result,
        torch.ones((variable,)),
        msg="RMSE of time mean between zero and 1 should be 1.",
    )


@pytest.mark.parametrize(*test_cases)
def test_time_and_global_mean(variable, time, lats, n_lon):
    x = torch.randn(variable, time, len(lats), n_lon)
    random_weights = torch.rand(len(lats), n_lon)

    result = fme.time_and_global_mean_bias(x, x.clone(), time_dim=1)
    torch.testing.assert_close(
        result,
        torch.zeros((variable,)),
        msg="Time and global mean bias between identical tensors should be zero.",
    )

    result = fme.time_and_global_mean_bias(
        torch.zeros(variable, time, len(lats), n_lon),
        torch.ones(variable, time, len(lats), n_lon),
        weights=random_weights,
        time_dim=1,
    )
    torch.testing.assert_close(
        result,
        torch.ones((variable,)),
        msg="Global mean bias between zero and 1 should be 1.",
    )


def test_gradient_magnitude():
    constant = torch.ones((5, 2, 4, 4))
    constant_grad_magnitude = fme.gradient_magnitude(constant, dim=(-2, -1))
    torch.testing.assert_close(constant_grad_magnitude, torch.zeros_like(constant))

    monotonic = torch.tile(torch.arange(4.0), (5, 2, 4, 1))
    monotonic_grad_magnitude = fme.gradient_magnitude(monotonic, dim=(-2, -1))
    torch.testing.assert_close(monotonic_grad_magnitude, torch.ones_like(monotonic))


def test_weighted_mean_gradient_magnitude():
    constant = torch.ones((5, 2, 4, 4))
    constant_grad_magnitude = fme.weighted_mean_gradient_magnitude(
        constant, dim=(-2, -1)
    )
    torch.testing.assert_close(constant_grad_magnitude, torch.zeros((5, 2)))

    monotonic = torch.tile(torch.arange(4.0), (5, 2, 4, 1))
    monotonic_grad_magnitude = fme.weighted_mean_gradient_magnitude(
        monotonic, dim=(-2, -1)
    )
    torch.testing.assert_close(monotonic_grad_magnitude, torch.ones((5, 2)))


def test_weighted_mean_gradient_magnitude_with_nans():
    all_nans = torch.tensor([float("nan"), float("nan")])
    all_nans_magnitude = fme.weighted_mean_gradient_magnitude(all_nans, dim=-1)
    assert np.isnan(all_nans_magnitude.item())

    constant_nan = torch.ones((5, 2, 4, 4))
    constant_nan[:, 0, 0, 0] = float("nan")

    weights = torch.ones((2, 4, 4))
    constant_grad_magnitude = fme.weighted_mean_gradient_magnitude(
        constant_nan, weights=weights, dim=(-2, -1)
    )
    torch.testing.assert_close(constant_grad_magnitude, torch.zeros((5, 2)))

    monotonic_nan = torch.tile(torch.arange(4.0), (5, 2, 4, 4))
    monotonic_nan[:, 0, 2, 2] = float("nan")
    monotonic_grad_magnitude = fme.weighted_mean_gradient_magnitude(
        monotonic_nan, dim=(-2, -1)
    )
    torch.testing.assert_close(monotonic_grad_magnitude, torch.ones((5, 2)))


def test_gradient_magnitude_percent_diff():
    constant = torch.ones((5, 2, 4, 4))
    self_percent_diff = fme.gradient_magnitude_percent_diff(
        constant, constant, dim=(-2, -1)
    )
    assert torch.all(torch.isnan(self_percent_diff))

    monotonic = torch.tile(torch.arange(4.0), (5, 2, 4, 1))
    percent_diff = fme.gradient_magnitude_percent_diff(
        monotonic, constant, dim=(-2, -1)
    )
    torch.testing.assert_close(percent_diff, -100 * torch.ones((5, 2)))


def test_dry_air_no_water():
    torch.manual_seed(0)
    nlat, nlon = 4, 8
    water = torch.zeros(nlat, nlon)
    pressure = torch.rand(nlat, nlon)
    dry_air = surface_pressure_due_to_dry_air(pressure, water)
    np.testing.assert_allclose(dry_air.cpu().numpy(), pressure.cpu().numpy())


def test_dry_air_all_water():
    torch.manual_seed(0)
    nlat, nlon = 4, 8
    pressure = torch.rand(nlat, nlon)
    water = pressure / GRAVITY
    dry_air = surface_pressure_due_to_dry_air(pressure, water)
    np.testing.assert_almost_equal(dry_air.cpu().numpy(), 0.0, decimal=6)


def test_net_surface_energy_flux():
    sfc_down_lw_radiative_flux = torch.tensor([100.0])
    sfc_up_lw_radiative_flux = torch.tensor([50.0])
    sfc_down_sw_radiative_flux = torch.tensor([25.0])
    sfc_up_sw_radiative_flux = torch.tensor([10.0])
    latent_heat_flux = torch.tensor([5.0])
    sensible_heat_flux = torch.tensor([2.5])
    expected_net_surface_energy_flux = torch.tensor([57.5])
    result = net_surface_energy_flux(
        sfc_down_lw_radiative_flux,
        sfc_up_lw_radiative_flux,
        sfc_down_sw_radiative_flux,
        sfc_up_sw_radiative_flux,
        latent_heat_flux,
        sensible_heat_flux,
    )
    torch.testing.assert_close(result, expected_net_surface_energy_flux)


@pytest.mark.parametrize(
    "bins, hist, pct, expected",
    [
        pytest.param(
            np.array([0, 1]), np.array([10]), 0.8, 0.8, id="single_bin_interpolation"
        ),
        pytest.param(
            np.array([0, 1, 2]),
            np.array([10, 10]),
            0.8,
            1.6,
            id="two_bin_interpolation",
        ),
        pytest.param(
            np.array([0, 1, 2]), np.array([10, 10]), 1.0, 2.0, id="two_bin_end_value"
        ),
    ],
)
def test_quantile(bins, hist, pct, expected):
    result = quantile(bins, hist, pct)
    np.testing.assert_allclose(result, expected)


def test_spherical_power_spectrum():
    nlat, nlon = 4, 8
    data = torch.rand(2, nlat, nlon, device=get_device())
    sht = torch_harmonics.RealSHT(nlat, nlon).to(get_device())
    spectrum = spherical_power_spectrum(data, sht)
    assert spectrum.shape == (2, nlat - 1)



import numpy as np
import torch
import xarray as xr

from fme.ace.aggregator.inference.spectrum import PairedSphericalPowerSpectrumAggregator
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import TensorMapping


class SpectrumAggregator:
    def __init__(
        self,
        gridded_operations: GriddedOperations,
        target_time: int = 1,
    ):
        self._wrapped = PairedSphericalPowerSpectrumAggregator(
            gridded_operations=gridded_operations,
            report_plot=False,
        )
        self._target_time = target_time

    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
        loss: torch.Tensor = torch.tensor(np.nan),
        i_time_start: int = 0,
    ):
        n_timesteps = next(iter(target_data.values())).shape[1]
        if (
            self._target_time >= i_time_start
            and self._target_time < i_time_start + n_timesteps
        ):
            i_time_target = self._target_time - i_time_start
            target_data = {
                key: value[:, i_time_target : i_time_target + 1, ...]
                for key, value in target_data.items()
            }
            gen_data = {
                key: value[:, i_time_target : i_time_target + 1, ...]
                for key, value in gen_data.items()
            }
            self._wrapped.record_batch(
                target_data,
                gen_data,
                target_data_norm,
                gen_data_norm,
                i_time_start,
            )

    def get_logs(self, label: str):
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        return self._wrapped.get_logs(label)

    def get_dataset(self) -> xr.Dataset:
        return self._wrapped.get_dataset()



import datetime
import logging
from collections.abc import Callable, MutableMapping

import torch

from fme.core.dataset.data_typing import VariableMetadata
from fme.core.ocean_data import HasOceanDepthIntegral, OceanData
from fme.core.typing_ import TensorDict

OceanDerivedVariableFunc = Callable[[OceanData, datetime.timedelta], torch.Tensor]

_OCEAN_DERIVED_VARIABLE_REGISTRY: MutableMapping[
    str, tuple[OceanDerivedVariableFunc, VariableMetadata, bool]
] = {}


def get_ocean_derived_variable_metadata() -> dict[str, VariableMetadata]:
    return {
        label: metadata
        for label, (_, metadata, _) in _OCEAN_DERIVED_VARIABLE_REGISTRY.items()
    }


def register(metadata: VariableMetadata, exists_ok: bool = False):
    def decorator(func: OceanDerivedVariableFunc):
        label = func.__name__
        if label in _OCEAN_DERIVED_VARIABLE_REGISTRY:
            raise ValueError(f"Function {label} has already been added to registry.")
        _OCEAN_DERIVED_VARIABLE_REGISTRY[label] = (func, metadata, exists_ok)
        return func

    return decorator


def _compute_ocean_derived_variable(
    data: TensorDict,
    depth_coordinate: HasOceanDepthIntegral | None,
    timestep: datetime.timedelta,
    label: str,
    derived_variable_func: OceanDerivedVariableFunc,
    forcing_data: TensorDict | None = None,
    exists_ok: bool = False,
) -> TensorDict:
    """Computes an ocean derived variable and adds it to the given data.

    By default the derived variable name must not already exist in the data,
    unless explicitly allowed with exists_ok=True.

    If any required input data are not available,
    the derived variable will not be computed.

    Args:
        data: dictionary of data to add the derived variable to.
        depth_coordinate: the depth coordinate.
        timestep: Timestep of the model.
        label: the name of the derived variable.
        derived_variable_func: derived variable function to compute.
        forcing_data: optional dictionary of forcing data needed for some derived
            variables. If necessary forcing inputs are missing, the derived
            variable will not be computed.
        exists_ok: Whether or not to allow the label to already exist in data,
            in which case a copy of the data TensorDict is returned with values
            unchanged.

    Returns:
        A new data dictionary with the derived variable added.
    """
    new_data = data.copy()
    if label in new_data:
        if exists_ok:
            return new_data
        raise ValueError(
            f"Variable {label} already exists. It is not permitted "
            "to have derived variables with same name as existing variables "
            "unless the derived variable is registered with exists_ok=True."
        )

    if forcing_data is not None:
        for key, value in forcing_data.items():
            if key not in data:
                data[key] = value

    ocean_data = OceanData(data, depth_coordinate)

    try:
        output = derived_variable_func(ocean_data, timestep)
    except KeyError as key_error:
        logging.debug(f"Could not compute {label} because {key_error} is missing")
    else:  # if no exception was raised
        new_data[label] = output
    return new_data


def compute_ocean_derived_quantities(
    data: TensorDict,
    depth_coordinate: HasOceanDepthIntegral | None,
    timestep: datetime.timedelta,
    forcing_data: TensorDict | None = None,
) -> TensorDict:
    """Computes all derived quantities from the given data."""
    for label in _OCEAN_DERIVED_VARIABLE_REGISTRY:
        func = _OCEAN_DERIVED_VARIABLE_REGISTRY[label][0]
        exists_ok = _OCEAN_DERIVED_VARIABLE_REGISTRY[label][2]
        data = _compute_ocean_derived_variable(
            data,
            depth_coordinate,
            timestep,
            label,
            func,
            forcing_data=forcing_data,
            exists_ok=exists_ok,
        )
    return data


@register(VariableMetadata("J/m**2", "Column-integrated ocean heat content"))
def ocean_heat_content(
    data: OceanData,
    timestep: datetime.timedelta,
) -> torch.Tensor:
    """Compute the column-integrated ocean heat content."""
    return data.ocean_heat_content


@register(VariableMetadata("[0-1]", "sea ice concentration"), exists_ok=True)
def sea_ice_fraction(
    data: OceanData,
    timestep: datetime.timedelta,
) -> torch.Tensor:
    """Compute the sea ice fraction."""
    return data.sea_ice_fraction



import copy
import dataclasses
import pathlib
import subprocess
import tempfile
import unittest.mock
from typing import Literal

import dacite
import numpy as np
import pytest
import torch
import xarray as xr
import yaml

import fme
from fme.ace.aggregator.inference.main import InferenceEvaluatorAggregatorConfig
from fme.ace.aggregator.one_step.main import OneStepAggregatorConfig
from fme.ace.data_loading.config import DataLoaderConfig
from fme.ace.data_loading.inference import (
    InferenceDataLoaderConfig,
    InferenceInitialConditionIndices,
)
from fme.ace.inference.data_writer.main import DataWriterConfig
from fme.ace.inference.evaluator import InferenceEvaluatorConfig
from fme.ace.inference.evaluator import main as inference_evaluator_main
from fme.ace.registry.test_hpx import (
    conv_next_block_config,
    decoder_config,
    down_sampling_block_config,
    encoder_config,
    output_layer_config,
    recurrent_block_config,
    up_sampling_block_config,
)
from fme.ace.stepper.single_module import StepperConfig
from fme.ace.testing import (
    DimSizes,
    MonthlyReferenceData,
    save_nd_netcdf,
    save_scalar_netcdf,
)
from fme.ace.train.train import build_trainer, prepare_directory
from fme.ace.train.train import main as train_main
from fme.ace.train.train_config import (
    InlineInferenceConfig,
    TrainBuilders,
    TrainConfig,
    WeatherEvaluationConfig,
)
from fme.core.coordinates import (
    HEALPixCoordinates,
    HorizontalCoordinates,
    LatLonCoordinates,
)
from fme.core.corrector.atmosphere import AtmosphereCorrectorConfig
from fme.core.dataset.config import XarrayDataConfig
from fme.core.generics.trainer import (
    _restore_checkpoint,
    count_parameters,
    epoch_checkpoint_enabled,
)
from fme.core.logging_utils import LoggingConfig
from fme.core.loss import WeightedMappingLossConfig
from fme.core.normalizer import NetworkAndLossNormalizationConfig, NormalizationConfig
from fme.core.ocean import OceanConfig
from fme.core.optimization import OptimizationConfig
from fme.core.registry.corrector import CorrectorSelector
from fme.core.registry.module import ModuleSelector
from fme.core.scheduler import SchedulerConfig
from fme.core.step.single_module import SingleModuleStepConfig
from fme.core.step.step import StepSelector
from fme.core.testing.model import compare_restored_parameters
from fme.core.testing.wandb import mock_wandb
from fme.core.typing_ import Slice

JOB_SUBMISSION_SCRIPT_PATH = (
    pathlib.PurePath(__file__).parent / "run-train-and-inference.sh"
)


def _get_test_yaml_files(
    *,
    train_data_path,
    valid_data_path,
    monthly_data_filename: pathlib.Path | None,
    results_dir,
    global_means_path,
    global_stds_path,
    in_variable_names,
    out_variable_names,
    mask_name,
    n_forward_steps=2,
    nettype="SphericalFourierNeuralOperatorNet",
    log_to_wandb=False,
    max_epochs=1,
    segment_epochs=1,
    inference_forward_steps=2,
    use_healpix=False,
    crps_training=False,
    save_per_epoch_diagnostics=False,
    log_validation_maps=False,
    skip_inline_inference=False,
    time_buffer=1,
):
    input_time_size = 1
    output_time_size = 1
    if use_healpix:
        in_channels = len(in_variable_names)
        out_channels = len(out_variable_names)
        prognostic_variables = min(
            out_channels, in_channels
        )  # how many variables in/out share.
        # in practice, we will need to compare variable names, since there
        # are some input-only and some output-only channels.
        # TODO: https://github.com/ai2cm/full-model/issues/1046
        n_constants = 0
        decoder_input_channels = 0  # was 1, to indicate insolation - now 0
        input_time_size = 1  # TODO: change to 2 (issue #1177)
        output_time_size = 1  # TODO: change to 4 (issue #1177)

        conv_next_block = conv_next_block_config(in_channels=in_channels)
        down_sampling_block = down_sampling_block_config()
        recurrent_block = recurrent_block_config()
        encoder = encoder_config(
            conv_next_block, down_sampling_block, n_channels=[16, 8, 4]
        )
        up_sampling_block = up_sampling_block_config()
        output_layer = output_layer_config()
        decoder = decoder_config(
            conv_next_block,
            up_sampling_block,
            output_layer,
            recurrent_block,
            n_channels=[4, 8, 16],
        )
        net_config = dict(
            encoder=encoder,
            decoder=decoder,
            prognostic_variables=prognostic_variables,
            n_constants=n_constants,
            decoder_input_channels=decoder_input_channels,
            input_time_size=input_time_size,
            output_time_size=output_time_size,
        )
        spatial_dimensions_str: Literal["healpix", "latlon"] = "healpix"
    elif nettype == "Samudra":
        net_config = dict(
            ch_width=[8, 16],
            dilation=[2, 4],
            n_layers=[1, 1],
        )
        spatial_dimensions_str = "latlon"
    elif nettype == "SphericalFourierNeuralOperatorNet":
        net_config = dict(
            num_layers=2,
            embed_dim=12,
        )
        spatial_dimensions_str = "latlon"
    elif nettype == "NoiseConditionedSFNO":
        net_config = dict(
            num_layers=2,
            embed_dim=12,
        )
        spatial_dimensions_str = "latlon"

    if nettype == "SphericalFourierNeuralOperatorNet":
        corrector_config: AtmosphereCorrectorConfig | CorrectorSelector = (
            CorrectorSelector(
                type="atmosphere_corrector",
                config=dataclasses.asdict(
                    AtmosphereCorrectorConfig(conserve_dry_air=True)
                ),
            )
        )
    else:
        corrector_config = AtmosphereCorrectorConfig()

    logging_config = LoggingConfig(
        log_to_screen=True,
        log_to_wandb=log_to_wandb,
        log_to_file=False,
        project="fme",
        entity="ai2cm",
    )
    if skip_inline_inference:
        inline_inference_config = None
        weather_evaluation_config = None
    else:
        inline_inference_config = InlineInferenceConfig(
            aggregator=InferenceEvaluatorAggregatorConfig(
                monthly_reference_data=(
                    str(monthly_data_filename)
                    if monthly_data_filename is not None
                    else None
                ),
            ),
            loader=InferenceDataLoaderConfig(
                dataset=XarrayDataConfig(
                    data_path=str(valid_data_path),
                    spatial_dimensions=spatial_dimensions_str,
                ),
                start_indices=InferenceInitialConditionIndices(
                    first=0,
                    n_initial_conditions=2,
                    interval=1,
                ),
            ),
            n_forward_steps=inference_forward_steps,
            forward_steps_in_memory=2,
        )
        weather_evaluation_config = WeatherEvaluationConfig(
            aggregator=InferenceEvaluatorAggregatorConfig(
                monthly_reference_data=(
                    str(monthly_data_filename)
                    if monthly_data_filename is not None
                    else None
                ),
            ),
            loader=InferenceDataLoaderConfig(
                dataset=XarrayDataConfig(
                    data_path=str(valid_data_path),
                    spatial_dimensions=spatial_dimensions_str,
                ),
                start_indices=InferenceInitialConditionIndices(
                    first=0,
                    n_initial_conditions=2,
                    interval=1,
                ),
            ),
            n_forward_steps=inference_forward_steps,
            forward_steps_in_memory=2,
        )

    train_config = TrainConfig(
        train_loader=DataLoaderConfig(
            dataset=XarrayDataConfig(
                data_path=str(train_data_path),
                spatial_dimensions=spatial_dimensions_str,
            ),
            batch_size=2,
            num_data_workers=0,
            time_buffer=time_buffer,
            sample_with_replacement=10,
        ),
        validation_loader=DataLoaderConfig(
            dataset=XarrayDataConfig(
                data_path=str(valid_data_path),
                spatial_dimensions=spatial_dimensions_str,
            ),
            batch_size=2,
            num_data_workers=0,
        ),
        optimization=OptimizationConfig(
            use_gradient_accumulation=True,
            optimizer_type="Adam",
            lr=0.001,
            kwargs=dict(weight_decay=0.01),
            scheduler=SchedulerConfig(
                type="CosineAnnealingLR",
                kwargs=dict(T_max=1),
            ),
        ),
        stepper=StepperConfig(
            loss=WeightedMappingLossConfig(type="MSE"),
            crps_training=crps_training,
            step=StepSelector(
                type="single_module",
                config=dataclasses.asdict(
                    SingleModuleStepConfig(
                        crps_training=crps_training,
                        in_names=in_variable_names,
                        out_names=out_variable_names,
                        normalization=NetworkAndLossNormalizationConfig(
                            network=NormalizationConfig(
                                global_means_path=str(global_means_path),
                                global_stds_path=str(global_stds_path),
                            ),
                            residual=NormalizationConfig(
                                global_means_path=str(global_means_path),
                                global_stds_path=str(global_stds_path),
                            ),
                        ),
                        builder=ModuleSelector(
                            type=nettype,
                            config=net_config,
                        ),
                        ocean=OceanConfig(
                            surface_temperature_name=in_variable_names[0],
                            ocean_fraction_name=mask_name,
                        ),
                        corrector=corrector_config,
                    )
                ),
            ),
        ),
        inference=inline_inference_config,
        weather_evaluation=weather_evaluation_config,
        n_forward_steps=n_forward_steps,
        max_epochs=max_epochs,
        segment_epochs=segment_epochs,
        save_checkpoint=True,
        logging=logging_config,
        experiment_dir=str(results_dir),
        save_per_epoch_diagnostics=save_per_epoch_diagnostics,
        validation_aggregator=OneStepAggregatorConfig(
            log_snapshots=log_validation_maps,
            log_mean_maps=log_validation_maps,
        ),
    )

    inference_config = InferenceEvaluatorConfig(
        experiment_dir=str(results_dir),
        n_forward_steps=6,
        forward_steps_in_memory=2,
        checkpoint_path=str(results_dir / "training_checkpoints" / "best_ckpt.tar"),
        data_writer=DataWriterConfig(
            save_prediction_files=True,
        ),
        aggregator=InferenceEvaluatorAggregatorConfig(
            log_video=True,
        ),
        logging=logging_config,
        loader=InferenceDataLoaderConfig(
            dataset=XarrayDataConfig(
                data_path=str(valid_data_path),
                spatial_dimensions=spatial_dimensions_str,
            ),
            start_indices=InferenceInitialConditionIndices(
                first=0,
                n_initial_conditions=2,
                interval=1,
            ),
        ),
    )

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".yaml") as f_train:
        f_train.write(yaml.dump(dataclasses.asdict(train_config)))

    with tempfile.NamedTemporaryFile(
        mode="w", delete=False, suffix=".yaml"
    ) as f_inference:
        f_inference.write(yaml.dump(dataclasses.asdict(inference_config)))

    return f_train.name, f_inference.name


def get_sizes(
    spatial_dims: HorizontalCoordinates = LatLonCoordinates(
        lon=torch.Tensor(np.arange(32)),
        lat=torch.Tensor(np.arange(16)),
    ),
    n_time=3,
    nz_interface=3,
) -> DimSizes:
    return DimSizes(
        n_time=n_time,
        horizontal=copy.deepcopy(spatial_dims.loaded_sizes),
        nz_interface=nz_interface,
    )


def _setup(
    path,
    nettype,
    log_to_wandb=False,
    max_epochs=1,
    segment_epochs=1,
    n_time=10,
    timestep_days=5,
    inference_forward_steps=2,
    use_healpix=False,
    save_per_epoch_diagnostics=False,
    crps_training=False,
    log_validation_maps=False,
    skip_inline_inference=False,
    time_buffer=1,
):
    if not path.exists():
        path.mkdir()
    seed = 0
    np.random.seed(seed)
    in_variable_names = [
        "PRESsfc",
        "specific_total_water_0",
        "specific_total_water_1",
        "surface_temperature",
        "baz",
    ]
    out_variable_names = [
        "PRESsfc",
        "specific_total_water_0",
        "specific_total_water_1",
        "surface_temperature",
    ]
    mask_name = "mask"
    all_variable_names = list(set(in_variable_names + out_variable_names))

    if use_healpix:
        hpx_coords = HEALPixCoordinates(
            face=torch.Tensor(np.arange(12)),
            width=torch.Tensor(np.arange(16)),
            height=torch.Tensor(np.arange(16)),
        )
        dim_sizes = get_sizes(spatial_dims=hpx_coords, n_time=n_time)
    else:
        dim_sizes = get_sizes(n_time=n_time)

    data_dir = path / "data"
    stats_dir = path / "stats"
    results_dir = path / "results"
    data_dir.mkdir()
    stats_dir.mkdir()
    results_dir.mkdir()
    save_nd_netcdf(
        data_dir / "data.nc",
        dim_sizes,
        variable_names=all_variable_names + [mask_name],
        timestep_days=timestep_days,
    )
    save_scalar_netcdf(
        stats_dir / "stats-mean.nc",
        variable_names=all_variable_names,
    )
    save_scalar_netcdf(
        stats_dir / "stats-stddev.nc",
        variable_names=all_variable_names,
    )

    monthly_dim_sizes: DimSizes
    if use_healpix:
        # monthly reference functionality not supported for HEALPix
        # see https://github.com/ai2cm/full-model/issues/1561
        monthly_data_filename = None
    else:
        monthly_dim_sizes = get_sizes(n_time=10 * 12, nz_interface=1)
        monthly_reference_data = MonthlyReferenceData(
            path=data_dir,
            names=out_variable_names,
            dim_sizes=monthly_dim_sizes,
            n_ensemble=3,
        )
        monthly_data_filename = monthly_reference_data.data_filename

    train_config_filename, inference_config_filename = _get_test_yaml_files(
        train_data_path=data_dir,
        valid_data_path=data_dir,
        monthly_data_filename=monthly_data_filename,
        results_dir=results_dir,
        global_means_path=stats_dir / "stats-mean.nc",
        global_stds_path=stats_dir / "stats-stddev.nc",
        in_variable_names=in_variable_names,
        out_variable_names=out_variable_names,
        mask_name=mask_name,
        nettype=nettype,
        log_to_wandb=log_to_wandb,
        max_epochs=max_epochs,
        segment_epochs=segment_epochs,
        inference_forward_steps=inference_forward_steps,
        use_healpix=use_healpix,
        crps_training=crps_training,
        save_per_epoch_diagnostics=save_per_epoch_diagnostics,
        log_validation_maps=log_validation_maps,
        skip_inline_inference=skip_inline_inference,
        time_buffer=time_buffer,
    )
    return train_config_filename, inference_config_filename


@pytest.mark.parametrize(
    "nettype, crps_training, log_validation_maps",
    [
        ("SphericalFourierNeuralOperatorNet", False, True),
        ("NoiseConditionedSFNO", True, False),
        ("HEALPixRecUNet", False, False),
        ("Samudra", False, False),
        ("NoiseConditionedSFNO", False, False),
    ],
)
def test_train_and_inference(
    tmp_path, nettype, crps_training, log_validation_maps: bool, very_fast_only: bool
):
    """Ensure that ACE training and subsequent standalone inference run without errors.

    Args:
        tmp_path: pytext fixture for temporary workspace.
        nettype: parameter indicating model architecture to use.
        very_fast_only: parameter indicating whether to skip slow tests.
    """
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")
    # need multi-year to cover annual aggregator
    train_config, inference_config = _setup(
        tmp_path,
        nettype,
        log_to_wandb=True,
        timestep_days=20,
        n_time=int(366 * 3 / 20 + 1),
        inference_forward_steps=int(366 * 3 / 20 / 2 - 1) * 2,  # must be even
        use_healpix=(nettype == "HEALPixRecUNet"),
        crps_training=crps_training,
        save_per_epoch_diagnostics=True,
        log_validation_maps=log_validation_maps,
    )
    # using pdb requires calling main functions directly
    with mock_wandb() as wandb:
        train_main(
            yaml_config=train_config,
        )
        wandb_logs = wandb.get_logs()

        for log in wandb_logs:
            # ensure inference time series is not logged
            assert "inference/mean/forecast_step" not in log

    validation_output_dir = tmp_path / "results" / "output" / "val" / "epoch_0001"
    assert validation_output_dir.exists()
    validation_diags = ["mean"]
    validation_map_diags = ["snapshot", "mean_map"]
    for diagnostic in validation_diags + validation_map_diags:
        diagnostic_output = validation_output_dir / f"{diagnostic}_diagnostics.nc"
        if diagnostic in validation_map_diags and not log_validation_maps:
            assert not diagnostic_output.exists()
        else:
            assert diagnostic_output.exists()
            ds = xr.open_dataset(diagnostic_output, decode_timedelta=False)
            assert len(ds) > 0

    inline_inference_output_dir = (
        tmp_path / "results" / "output" / "inference" / "epoch_0001"
    )
    assert inline_inference_output_dir.exists()
    for diagnostic in (
        "mean_step_20",
        "time_mean",
        "time_mean_norm",
        "annual",
    ):
        diagnostic_output = inline_inference_output_dir / f"{diagnostic}_diagnostics.nc"
        assert diagnostic_output.exists()
        ds = xr.open_dataset(diagnostic_output, decode_timedelta=False)
        assert len(ds) > 0

    # inference should not require stats files
    (tmp_path / "stats" / "stats-mean.nc").unlink()
    (tmp_path / "stats" / "stats-stddev.nc").unlink()

    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        inference_evaluator_main(yaml_config=inference_config)
        inference_logs = wandb.get_logs()

    prediction_output_path = tmp_path / "results" / "autoregressive_predictions.nc"
    best_checkpoint_path = (
        tmp_path / "results" / "training_checkpoints" / "best_ckpt.tar"
    )
    best_inference_checkpoint_path = (
        tmp_path / "results" / "training_checkpoints" / "best_inference_ckpt.tar"
    )
    assert best_checkpoint_path.exists()
    checkpoint_training_history = torch.load(best_checkpoint_path, weights_only=False)[
        "stepper"
    ].get("training_history")
    assert checkpoint_training_history is not None
    assert len(checkpoint_training_history) == 1
    assert "git_sha" in checkpoint_training_history[0].keys()
    assert best_inference_checkpoint_path.exists()
    n_ic_timesteps = 1
    n_forward_steps = 6
    n_summary_steps = 1
    assert len(inference_logs) == n_ic_timesteps + n_forward_steps + n_summary_steps
    assert prediction_output_path.exists()
    ds_prediction = xr.open_dataset(prediction_output_path, decode_timedelta=False)
    assert np.sum(np.isnan(ds_prediction["PRESsfc"].values)) == 0
    assert np.sum(np.isnan(ds_prediction["specific_total_water_0"].values)) == 0
    assert np.sum(np.isnan(ds_prediction["specific_total_water_1"].values)) == 0
    assert np.sum(np.isnan(ds_prediction["total_water_path"].values)) == 0
    ds_target = xr.open_dataset(
        tmp_path / "results" / "autoregressive_target.nc", decode_timedelta=False
    )
    assert np.sum(np.isnan(ds_target["baz"].values)) == 0


@pytest.mark.parametrize("nettype", ["SphericalFourierNeuralOperatorNet"])
def test_resume(tmp_path, nettype, very_fast_only: bool):
    """Make sure the training is resumed from a checkpoint when restarted."""
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")

    mock = unittest.mock.MagicMock(side_effect=_restore_checkpoint)
    with unittest.mock.patch("fme.core.generics.trainer._restore_checkpoint", new=mock):
        train_config, _ = _setup(
            tmp_path, nettype, log_to_wandb=True, max_epochs=2, segment_epochs=1
        )
        with mock_wandb() as wandb:
            train_main(yaml_config=train_config)
            assert (
                min([val["epoch"] for val in wandb.get_logs() if "epoch" in val]) == 1
            )
            assert (
                max([val["epoch"] for val in wandb.get_logs() if "epoch" in val]) == 1
            )
            assert not mock.called
            # need to persist the id since mock_wandb doesn't
            id = wandb.get_id()
        with mock_wandb() as wandb:
            # set the id so that we can check it matches what's in the experiment dir
            wandb.set_id(id)
            train_main(yaml_config=train_config)
            mock.assert_called()
            assert (
                min([val["epoch"] for val in wandb.get_logs() if "epoch" in val]) == 2
            )
            assert (
                max([val["epoch"] for val in wandb.get_logs() if "epoch" in val]) == 2
            )


def test_restore_checkpoint(tmp_path, very_fast_only: bool):
    """Test that restoring a checkpoint works."""
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")

    config_path, _ = _setup(tmp_path, "SphericalFourierNeuralOperatorNet")
    with open(config_path) as f:
        config_dict = yaml.safe_load(f)
    config = dacite.from_dict(
        data_class=TrainConfig, data=config_dict, config=dacite.Config(strict=True)
    )
    prepare_directory(config.experiment_dir, config_dict)

    builders = TrainBuilders(config)
    trainer1 = build_trainer(builders, config)

    # run one epoch
    trainer1.train_one_epoch()
    trainer1.save_all_checkpoints(0.1, 0.2)

    # reload and check model parameters and optimizer state
    trainer2 = build_trainer(builders, config)
    _restore_checkpoint(
        trainer2,
        trainer1.paths.latest_checkpoint_path,
        trainer1.paths.ema_checkpoint_path,
    )

    compare_restored_parameters(
        trainer1.stepper.modules.parameters(),
        trainer2.stepper.modules.parameters(),
        trainer1.optimization.optimizer,
        trainer2.optimization.optimizer,
    )


@pytest.mark.parametrize("nettype", ["SphericalFourierNeuralOperatorNet"])
@pytest.mark.skipif(
    fme.get_device().type == "mps", reason="MPS does not support multi-device training."
)
def test_resume_two_workers(tmp_path, nettype, skip_slow: bool, tmpdir: pathlib.Path):
    """Make sure the training is resumed from a checkpoint when restarted, using
    torchrun with NPROC_PER_NODE set to 2."""
    if skip_slow:
        # script is slow as everything is re-imported when it runs
        pytest.skip("Skipping slow tests")
    train_config, inference_config = _setup(tmp_path, nettype)
    subprocess_args = [
        JOB_SUBMISSION_SCRIPT_PATH,
        train_config,
        inference_config,
        "2",  # this makes the training run on two GPUs
    ]
    initial_process = subprocess.run(subprocess_args, cwd=tmpdir)
    initial_process.check_returncode()
    resume_subprocess_args = [
        "torchrun",
        "--nproc_per_node",
        "2",
        "-m",
        "fme.ace.train",
        train_config,
    ]
    resume_process = subprocess.run(resume_subprocess_args, cwd=tmpdir)
    resume_process.check_returncode()


def _create_fine_tuning_config(path_to_train_config_yaml: str, path_to_checkpoint: str):
    # TODO(gideond) rename to "overwrite" or something of that nature
    with open(path_to_train_config_yaml) as config_file:
        config_data = yaml.safe_load(config_file)
        config_data["stepper"] = {"checkpoint_path": path_to_checkpoint}
        current_experiment_dir = config_data["experiment_dir"]
        new_experiment_dir = pathlib.Path(current_experiment_dir) / "fine_tuning"
        config_data["experiment_dir"] = str(new_experiment_dir)
        with tempfile.NamedTemporaryFile(
            mode="w", delete=False, suffix=".yaml"
        ) as new_config_file:
            new_config_file.write(yaml.dump(config_data))

    return new_config_file.name, new_experiment_dir


@pytest.mark.parametrize("nettype", ["SphericalFourierNeuralOperatorNet"])
def test_fine_tuning(tmp_path, nettype, very_fast_only: bool):
    """Check that fine tuning config runs without errors."""
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")
    train_config, _ = _setup(tmp_path, nettype)

    train_main(yaml_config=train_config)

    results_dir = tmp_path / "results"
    ckpt = f"{results_dir}/training_checkpoints/best_ckpt.tar"

    fine_tuning_config, new_results_dir = _create_fine_tuning_config(train_config, ckpt)

    train_main(yaml_config=fine_tuning_config)
    assert (new_results_dir / "training_checkpoints" / "ckpt.tar").exists()


def _create_copy_weights_after_batch_config(
    path_to_train_config_yaml: str, path_to_checkpoint: str, experiment_dir: str
):
    with open(path_to_train_config_yaml) as config_file:
        config_data = yaml.safe_load(config_file)
        config_data["stepper"]["parameter_init"] = {"weights_path": path_to_checkpoint}
        config_data["copy_weights_after_batch"] = [{"include": ["*"], "exclude": []}]
        config_data["experiment_dir"] = experiment_dir
        with tempfile.NamedTemporaryFile(
            mode="w", delete=False, suffix=".yaml"
        ) as new_config_file:
            new_config_file.write(yaml.dump(config_data))

    return new_config_file.name


@pytest.mark.parametrize("nettype", ["SphericalFourierNeuralOperatorNet"])
def test_copy_weights_after_batch(tmp_path, nettype, skip_slow: bool):
    """Check that fine tuning config using copy_weights_after_batch
    runs without errors."""
    if skip_slow:
        pytest.skip("Skipping slow tests")

    train_config, _ = _setup(tmp_path, nettype)

    train_main(
        yaml_config=train_config,
    )

    results_dir = tmp_path / "results"
    ckpt = f"{results_dir}/training_checkpoints/best_ckpt.tar"

    fine_tuning_config = _create_copy_weights_after_batch_config(
        train_config, ckpt, experiment_dir=str(tmp_path / "fine_tuning_dir")
    )
    train_main(yaml_config=fine_tuning_config)


@pytest.mark.parametrize(
    "checkpoint_save_epochs,expected_save_epochs",
    [(None, []), (Slice(start=-2), [2, 3]), (Slice(step=2), [0, 2])],
)
def test_epoch_checkpoint_enabled(checkpoint_save_epochs, expected_save_epochs):
    max_epochs = 4
    for i in range(max_epochs):
        if i in expected_save_epochs:
            assert epoch_checkpoint_enabled(i, max_epochs, checkpoint_save_epochs)
        else:
            assert not epoch_checkpoint_enabled(i, max_epochs, checkpoint_save_epochs)


@pytest.mark.parametrize(
    "module_list,expected_num_parameters",
    [
        (torch.nn.ModuleList([torch.nn.Linear(10, 5), torch.nn.Linear(5, 2)]), 67),
        (torch.nn.ModuleList([]), 0),
    ],
)
def test_count_parameters(module_list, expected_num_parameters):
    num_parameters = count_parameters(module_list)
    assert num_parameters == expected_num_parameters


def test_train_without_inline_inference(tmp_path, very_fast_only: bool):
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")
    nettype = "SphericalFourierNeuralOperatorNet"
    crps_training = False
    log_validation_maps = False
    train_config, inference_config = _setup(
        tmp_path,
        nettype,
        log_to_wandb=True,
        timestep_days=20,
        n_time=int(366 * 3 / 20 + 1),
        inference_forward_steps=int(366 * 3 / 20 / 2 - 1) * 2,  # must be even
        use_healpix=(nettype == "HEALPixRecUNet"),
        crps_training=crps_training,
        save_per_epoch_diagnostics=True,
        log_validation_maps=log_validation_maps,
        skip_inline_inference=True,
        time_buffer=2,
    )
    with mock_wandb() as wandb:
        train_main(
            yaml_config=train_config,
        )
        wandb_logs = wandb.get_logs()
    assert np.isinf(wandb_logs[-1]["best_inference_error"])
    assert not any("inference/" in key for key in wandb_logs[-1])



import datetime

import cftime
import numpy as np
import pytest
import torch
import xarray as xr

from fme.ace.inference.data_writer.dataset_metadata import DatasetMetadata
from fme.ace.inference.data_writer.monthly import (
    MonthlyDataWriter,
    add_data,
    find_boundary,
    get_days_since_reference,
    months_for_timesteps,
)
from fme.core.dataset.data_typing import VariableMetadata

TIMESTEP = datetime.timedelta(hours=6)


@pytest.mark.parametrize(
    "window_size",
    [
        1,
        2,
    ],
)
@pytest.mark.parametrize(
    "n_writes",
    [1, 2],
)
def test_monthly_data_writer(tmpdir, window_size: int, n_writes: int):
    n_samples = 2
    n_lat, n_lon = 8, 16
    writer = MonthlyDataWriter(
        path=str(tmpdir),
        label="predictions",
        n_samples=n_samples,
        n_months=24,
        save_names=None,
        variable_metadata={"x": VariableMetadata(units="m", long_name="x_name")},
        coords={},
        dataset_metadata=DatasetMetadata(source={"inference_version": "1.0"}),
    )
    month_values = []
    for year in range(2020, 2022):
        for month in range(1, 13):
            x = torch.rand(n_samples, 1, n_lat, n_lon) + 1.0
            month_values.append(x)
            # repeat x along axis 1 to simulate a window_size > 1
            x_window = torch.cat([x] * window_size, dim=1)
            month_data = {"x": x_window}
            initial_time = cftime.DatetimeProlepticGregorian(year, month, 1, 0, 0, 0)
            for i_write in range(n_writes):
                time = xr.DataArray(
                    [
                        [
                            initial_time + datetime.timedelta(hours=6 * i_write)
                            for _ in range(window_size)
                        ]
                        for _ in range(n_samples)
                    ],
                    dims=["sample", "time"],
                )
                assert time.shape == (n_samples, window_size)
                writer.append_batch(data=month_data, start_timestep=0, batch_time=time)
    writer.flush()
    written = xr.open_dataset(
        str(tmpdir / "monthly_mean_predictions.nc"), decode_timedelta=False
    )
    assert written["x"].shape == (n_samples, 24, n_lat, n_lon)
    assert np.sum(written["x"].values != 0) > 0, "No non-zero values written"
    assert (
        np.sum(written["x"].values == 0.0) == 0
    ), "Some values are zero (were not added to)"
    np.testing.assert_array_equal(written["counts"].values, window_size * n_writes)
    np.testing.assert_allclose(
        written["x"],
        torch.cat(month_values, dim=1).cpu().numpy(),
    )
    assert "counts" in written.coords
    assert "counts" in written.x.coords
    assert "counts" in written.valid_time.coords
    assert written.attrs["title"] == "ACE monthly predictions data file"
    assert written.attrs["source.inference_version"] == "1.0"


@pytest.mark.parametrize(
    "n_timesteps, min_expected",
    [
        (1, 1),
        (2, 2),  # 2 timesteps can cross a month boundary
        (3, 2),  # 3 timesteps can't cross an additional boundary
        (4 * 28, 2),  # 28 days can only cross one month boundary
        (4 * 28 + 1, 3),  # 29 days can cross two month boundaries
    ],
)
def test_months_for_timesteps(n_timesteps: int, min_expected: int):
    assert months_for_timesteps(n_timesteps, TIMESTEP) >= min_expected


@pytest.mark.parametrize("num_years", [2, 500])
@pytest.mark.parametrize("calendar", ["proleptic_gregorian", "noleap"])
def test_get_days_since_reference(num_years, calendar):
    first_year = 2020
    final_year = first_year + num_years - 1
    years = np.array([i for i in range(first_year, final_year + 1)])
    months = np.zeros((num_years,), dtype=int)
    # For last year set month to 1
    months[-1] = 1
    if calendar == "proleptic_gregorian":
        reference_date = cftime.DatetimeProlepticGregorian(2020, 1, 1)
    else:
        reference_date = cftime.DatetimeNoLeap(2020, 1, 1)
    n_months = 3
    days = get_days_since_reference(years, months, reference_date, n_months, calendar)
    assert days.shape == (num_years, 3)
    # 2020 is a leap year in proleptic_gregorian
    if calendar == "proleptic_gregorian":
        assert days[0, 0] == 0
        assert days[0, 1] == 31
        assert days[0, 2] == 31 + 29
        if num_years == 2:
            assert days[1, 0] == 366 + 31
            assert days[1, 1] == 366 + 31 + 28
            assert days[1, 2] == 366 + 31 + 28 + 31
        if num_years == 500:
            # 121 is number of leap days
            assert days[499, 0] == 182135 + 121 + 31
            assert days[499, 1] == 182135 + 121 + 31 + 28
    if calendar == "noleap":
        assert days[0, 0] == 0
        assert days[0, 1] == 31
        assert days[0, 2] == 31 + 28
        if num_years == 2:
            assert days[1, 0] == 365 + 31
            assert days[1, 1] == 365 + 31 + 28
            assert days[1, 2] == 365 + 31 + 28 + 31
        if num_years == 500:
            assert days[499, 0] == 182135 + 31
            assert days[499, 1] == 182135 + 31 + 28


@pytest.mark.parametrize(
    "month_array, expected",
    [
        pytest.param([1, 2, 3, 4, 5, 6], 1, id="linear"),
        pytest.param([1, 1, 2], 2, id="after two steps"),
        pytest.param([1], 1, id="one value"),
        pytest.param([1, 1, 1, 1], 4, id="all the same"),
        pytest.param([0] * 50 + [1] * (23), 50, id="long array case"),
    ],
)
def test_find_boundary(month_array, expected):
    assert (
        find_boundary(np.asarray(month_array), start_month=month_array[0]) == expected
    )


def test_add_data_one_first_month():
    target = np.zeros((2, 3))
    target_start_counts = np.zeros((2, 3), dtype=np.int32)
    source = np.ones((2, 5))
    months_elapsed = np.zeros((2, 5), dtype=np.int32)
    expected = np.zeros((2, 3))
    expected[:, 0] = 1

    add_data(
        target=target,
        target_start_counts=target_start_counts,
        source=source,
        months_elapsed=months_elapsed,
    )
    np.testing.assert_array_equal(target, expected)
    np.testing.assert_array_equal(target_start_counts, 0)


def test_add_data_one_first_month_averaging():
    target = np.zeros((2, 3))
    target[0, 0] = 2.0
    target_start_counts = np.zeros((2, 3), dtype=np.int32)
    target_start_counts[0, 0] = 1
    source = np.ones((2, 5))
    months_elapsed = np.zeros((2, 5), dtype=np.int32)
    expected = np.zeros((2, 3))
    expected[0, 0] = (2 + 5) / 6
    expected[1, 0] = 1

    add_data(
        target=target,
        target_start_counts=target_start_counts,
        source=source,
        months_elapsed=months_elapsed,
    )
    np.testing.assert_array_equal(target, expected)


def test_add_data_one_later_month():
    target = np.zeros((2, 4))
    target_start_counts = np.zeros((2, 4), dtype=np.int32)
    source = np.ones((2, 5))
    months_elapsed = np.zeros((2, 5), dtype=np.int32) + 2
    expected = np.zeros((2, 4))
    expected[:, 2] = 1

    add_data(
        target=target,
        target_start_counts=target_start_counts,
        source=source,
        months_elapsed=months_elapsed,
    )
    np.testing.assert_array_equal(target, expected)


def test_add_data_two_later_months():
    target = np.zeros((2, 4))
    target_start_counts = np.zeros((2, 4), dtype=np.int32)
    source = np.ones((2, 5))
    months_elapsed = np.zeros((2, 5), dtype=np.int32) + 2
    months_elapsed[0, 2:] = 3
    months_elapsed[1, 3:] = 3
    expected = np.zeros((2, 4))
    expected[0, 2] = 1
    expected[0, 3] = 1
    expected[1, 2] = 1
    expected[1, 3] = 1

    add_data(
        target=target,
        target_start_counts=target_start_counts,
        source=source,
        months_elapsed=months_elapsed,
    )
    np.testing.assert_array_equal(target, expected)



import abc
from typing import Any, Generic, TypeVar

PS = TypeVar("PS", contravariant=True)  # prognostic state
SD = TypeVar("SD", contravariant=True)  # stepped data


class WriterABC(abc.ABC, Generic[PS, SD]):
    @abc.abstractmethod
    def write(self, data: PS, filename: str):
        """Eagerly write data to a file at filename."""
        ...

    @abc.abstractmethod
    def append_batch(
        self,
        batch: SD,
    ):
        """
        Append a batch of data to the output file(s).

        Args:
            batch: Data to be written.
        """
        ...


class NullDataWriter(WriterABC[Any, Any]):
    """
    Null pattern for DataWriter, which does nothing.
    """

    def __init__(self):
        pass

    def append_batch(
        self,
        batch: Any,
    ):
        pass

    def flush(self):
        pass

    def write(self, data: Any, filename: str):
        pass



import torch


class CappedGELU(torch.nn.Module):
    """
    Implements a GeLU with capped maximum value.
    """

    def __init__(self, cap_value: float = 1.0, **kwargs):
        """
        :param cap_value: float: value at which to clip activation
        :param kwargs: passed to torch.nn.LeadyReLU
        """
        super().__init__()
        self.add_module("gelu", torch.nn.GELU(**kwargs))
        # self.cap = torch.tensor(cap_value, dtype=torch.float32)
        self.register_buffer("cap", torch.tensor(cap_value, dtype=torch.float32))

    def forward(self, inputs: torch.Tensor):
        x = self.gelu(inputs)
        # Convert cap to a scalar value for clamping (ignores grad)
        cap_value = self.cap.item()
        x = torch.clamp(x, max=cap_value)
        return x



import argparse
import dataclasses
import shutil
import tempfile
import warnings
from typing import Callable, List, Optional, Sequence, Tuple

import dacite
import dask.diagnostics
import fsspec
import numpy as np
import xarray as xr
import yaml
from scipy.optimize import curve_fit

from fme.core import metrics


@dataclasses.dataclass
class DataConfig:
    paths: List[str]
    stats_path: str
    years_per_ensemble: int
    is_amip: bool
    start_date: Optional[str] = None
    end_date: Optional[str] = None

    def get_dataset(self) -> xr.Dataset:
        datasets = [xr.open_zarr(path) for path in self.paths]
        ds = xr.concat(datasets, "sample")
        ds = ds.sel(time=slice(self.start_date, self.end_date))
        return ds


def copy(source: str, destination: str):
    """Copy between any two 'filesystems'. Do not use for large files.

    Args:
        source: Path to source file/object.
        destination: Path to destination.
    """
    with fsspec.open(source) as f_source:
        with fsspec.open(destination, "wb") as f_destination:
            shutil.copyfileobj(f_source, f_destination)


def get_annual(ds: xr.Dataset):
    """
    Resamples a dataset into annual means.
    """
    samples = [ds.isel(sample=i) for i in range(len(ds.sample))]
    annual_samples = [s.groupby(s.time.dt.year).mean() for s in samples]
    return xr.concat(annual_samples, "sample")


def get_stats(
    annual: xr.Dataset,
    years_per_ensemble: int,
    ensemble_members: int,
    window_size: int,
    area: xr.DataArray,
    amip: bool,
):
    """
    Gets the mean and standard deviation of the pattern bias implied by windows
    of a given number of years, from a dataset of annual means.
    """
    n_windows_per_sample = years_per_ensemble // window_size
    rmses = []
    if not amip:
        mean = annual.mean(["sample", "year"])
        annual_bias = annual - mean
    else:
        annual_bias = annual  # bias calculated on each window separately
    for i_w in range(n_windows_per_sample):
        window = annual_bias.isel(
            year=range(i_w * window_size, (i_w + 1) * window_size)
        )
        if amip:
            window = window - window.mean(["sample", "year"])
        bias_maps = window.mean("year")
        rmse = (bias_maps**2).weighted(area).mean(dim=["grid_yt", "grid_xt"]) ** 0.5
        if amip:
            rmse *= (ensemble_members / (ensemble_members - 1.0)) ** 0.5
        else:
            n_windows = n_windows_per_sample * ensemble_members
            rmse *= (n_windows / (n_windows - 1)) ** 0.5
        rmses.append(rmse)
    rmses_da = xr.concat(rmses, "window")
    mean_rmses = rmses_da.mean()
    stdev_rmses = rmses_da.std()
    return mean_rmses, stdev_rmses


def combine_window_stats(
    window_sizes: Sequence[int],
    years_per_ensemble: int,
    ensemble_members: int,
    func,
    annual: xr.Dataset,
    area: xr.DataArray,
    amip: bool,
):
    """
    Gets the mean and standard deviation of the pattern bias implied by windows
    of multiple different year lengths, from a dataset of annual means.
    """
    window_sizes = list(window_sizes)
    means = []
    stdevs = []
    for window_size in window_sizes:
        m, s = func(
            annual, years_per_ensemble, ensemble_members, window_size, area, amip
        )
        means.append(m)
        stdevs.append(s)
    m = xr.concat(means, "window_size")
    m = m.assign_coords({"window_size": window_sizes})
    s = xr.concat(stdevs, "window_size")
    s = s.assign_coords({"window_size": window_sizes})
    return m, s


def fit(x: np.ndarray, y: np.ndarray) -> Callable[[np.ndarray], np.ndarray]:
    """
    Fits given data to a function of the form y = c / sqrt(x) + b.
    """

    def model(x, c, b):
        return c * 1 / np.sqrt(x) + b

    popt, _ = curve_fit(model, x, y)

    c, b = popt

    def fitted_model(x):
        return model(x, c, b)

    return fitted_model


def get_output_datasets(
    dataset: xr.Dataset,
    years_per_ensemble: int,
    amip: bool,
) -> Tuple[xr.Dataset, xr.Dataset, xr.Dataset]:
    lat = dataset["grid_yt"]
    area = xr.DataArray(
        metrics.spherical_area_weights(lat.values, dataset.sizes["grid_xt"]),
        dims=["grid_yt", "grid_xt"],
    )
    annual = get_annual(dataset)
    available_years = annual.sizes["year"]
    if available_years != years_per_ensemble:
        warnings.warn(
            f"There are {available_years} years of data available, "
            f"but received value {years_per_ensemble} for years_per_ensemble."
        )
    with dask.diagnostics.ProgressBar():
        with dask.config.set(scheduler="processes"):
            annual = annual.isel(year=range(0, years_per_ensemble)).load()
    m, s = combine_window_stats(
        [1, 2, 5, 10],
        years_per_ensemble,
        len(dataset.sample),
        get_stats,
        annual,
        area,
        amip,
    )
    m = m.load()
    s = s.load()
    return m, s, annual


def main(
    dataset: xr.Dataset,
    years_per_ensemble: int,
    amip: bool,
    stats_path: str,
):
    m, s, annual_means = get_output_datasets(
        dataset=dataset,
        years_per_ensemble=years_per_ensemble,
        amip=amip,
    )
    if stats_path.endswith("/"):
        stats_path = stats_path[:-1]

    with tempfile.TemporaryDirectory() as tmpdir:
        m.to_netcdf(tmpdir + "/rmse_means.nc")
        s.to_netcdf(tmpdir + "/rmse_stdevs.nc")
        annual_means.to_netcdf(tmpdir + "/annual_means.nc")
        copy(tmpdir + "/rmse_means.nc", stats_path + "/rmse_means.nc")
        copy(tmpdir + "/rmse_stdevs.nc", stats_path + "/rmse_stdevs.nc")
        copy(tmpdir + "/annual_means.nc", stats_path + "/annual_means.nc")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("data_config", type=str)
    args = parser.parse_args()

    with open(args.data_config, "r") as f:
        data_config = dacite.from_dict(DataConfig, yaml.safe_load(f))

    main(
        dataset=data_config.get_dataset(),
        years_per_ensemble=data_config.years_per_ensemble,
        amip=data_config.is_amip,
        stats_path=data_config.stats_path,
    )



import datetime

import numpy as np
import xarray as xr

from fme.core import logging_utils
from fme.core.dataset.config import XarrayDataConfig
from fme.core.testing.wandb import mock_wandb

from .benchmark import BenchmarkConfig, benchmark
from .config import ConcatDatasetConfig, DataLoaderConfig


def get_test_data(path):
    """Create a test dataset for benchmarking."""
    n_lat = 2
    n_lon = 4
    n_time = 10
    ds = xr.Dataset(
        {
            "a": (("time", "lat", "lon"), np.random.rand(n_time, n_lat, n_lon)),
            "b": (("time", "lat", "lon"), np.random.rand(n_time, n_lat, n_lon)),
        },
        coords={
            "time": (
                "time",
                [datetime.datetime(2023, 1, i + 1) for i in range(n_time)],
            ),
            "lat": ("lat", np.arange(n_lat)),
            "lon": ("lon", np.arange(n_lon)),
        },
    )
    ds.to_netcdf(path)
    return ds


def test_benchmark(tmp_path):
    """Ensure that the benchmark runs without error."""
    ds = get_test_data(tmp_path / "test.nc")
    dataset_config = XarrayDataConfig(data_path=str(tmp_path))
    loader_config = DataLoaderConfig(
        dataset=ConcatDatasetConfig(concat=[dataset_config]), batch_size=4
    )
    logging_config = logging_utils.LoggingConfig(project="test", entity="test")
    config = BenchmarkConfig(
        loader=loader_config,
        logging=logging_config,
        names=list(ds.data_vars),
        n_timesteps=3,
        sleep=0,
    )
    with mock_wandb() as wandb:  # noqa: F841
        benchmark(config)



import datetime
import pathlib

import cftime
import matplotlib.pyplot as plt
import numpy as np
import pytest
import torch
import xarray as xr

import fme
from fme.ace.aggregator.inference.annual import (
    GlobalMeanAnnualAggregator,
    PairedGlobalMeanAnnualAggregator,
)
from fme.ace.testing import DimSizes, MonthlyReferenceData
from fme.core.coordinates import DimSize
from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations
from fme.core.mask_provider import MaskProvider
from fme.core.testing import mock_distributed

TIMESTEP = datetime.timedelta(hours=6)


def test_paired_annual_aggregator(tmpdir):
    n_lat = 16
    n_lon = 32
    # need to have two actual full years of data for plotting to get exercised
    n_sample = 2
    n_time = 365 * 4 * 2
    area_weights = torch.ones(n_lat, n_lon).to(fme.get_device())
    names = ["a"]
    horizontal = [DimSize("grid_yt", n_lat), DimSize("grid_xt", n_lon)]
    monthly_reference_data = MonthlyReferenceData(
        path=pathlib.Path(tmpdir),
        names=names,
        dim_sizes=DimSizes(
            n_time=48,
            horizontal=horizontal,
            nz_interface=1,
        ),
        n_ensemble=3,
    )
    monthly_ds = xr.open_dataset(
        monthly_reference_data.data_filename,
        decode_timedelta=False,
    )
    agg = PairedGlobalMeanAnnualAggregator(
        ops=LatLonOperations(area_weights),
        timestep=TIMESTEP,
        monthly_reference_data=monthly_ds,
    )
    target_data = {
        "a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())
    }
    gen_data = {"a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())}
    time = xr.DataArray(
        [
            [
                (
                    cftime.DatetimeProlepticGregorian(2000, 1, 1)
                    + i * datetime.timedelta(hours=6)
                )
                for i in range(n_time)
            ]
            for _ in range(n_sample)
        ],
        dims=["sample", "time"],
    )
    agg.record_batch(time, target_data, gen_data)
    logs = agg.get_logs(label="test")
    assert len(logs) > 0
    assert "test/a" in logs
    assert isinstance(logs["test/a"], plt.Figure)
    assert "test/r2/a_target" in logs


def test_paired_annual_aggregator_with_nans(tmpdir):
    torch.manual_seed(0)
    n_lat = 16
    n_lon = 32
    # need to have two actual full years of data for plotting to get exercised
    n_sample = 2
    n_years = 5
    n_time = 365 * 4 * n_years
    area_weights = torch.ones(n_lat, n_lon).to(fme.get_device())
    names = ["a"]
    horizontal = [DimSize("grid_yt", n_lat), DimSize("grid_xt", n_lon)]
    monthly_reference_data = MonthlyReferenceData(
        path=pathlib.Path(tmpdir),
        names=names,
        dim_sizes=DimSizes(
            n_time=n_years * 12,
            horizontal=horizontal,
            nz_interface=1,
        ),
        n_ensemble=3,
    )
    monthly_ds = xr.open_dataset(
        monthly_reference_data.data_filename,
        decode_timedelta=False,
    )
    mask = np.ones((n_lat, n_lon))
    mask[1, 1] = 0
    monthly_ds["a"] = monthly_ds["a"].where(mask > 0)
    mask_provider = MaskProvider({"mask_a": torch.tensor(mask)}).to(get_device())
    agg = PairedGlobalMeanAnnualAggregator(
        ops=LatLonOperations(area_weights, mask_provider),
        timestep=TIMESTEP,
        monthly_reference_data=monthly_ds,
    )
    target_data = {
        "a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())
    }
    gen_data = {"a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())}
    target_data["a"][:, :, 1, 1] = float("nan")
    gen_data["a"][:, :, 1, 1] = float("nan")
    time = xr.DataArray(
        [
            [
                (
                    cftime.DatetimeProlepticGregorian(2000, 1, 1)
                    + i * datetime.timedelta(hours=6)
                )
                for i in range(n_time)
            ]
            for _ in range(n_sample)
        ],
        dims=["sample", "time"],
    )
    agg.record_batch(time, target_data, gen_data)
    annual_mean_series = agg.get_dataset()["a"].values
    expected_series = (
        xr.DataArray(
            np.concatenate(
                [
                    target_data["a"].unsqueeze(dim=0).cpu().numpy(),
                    gen_data["a"].unsqueeze(dim=0).cpu().numpy(),
                ]
            ),
            dims={
                "source": ["target", "prediction"],
                "sample": np.arange(n_sample),
                "time": time,
                "lat": np.arange(n_lat),
                "lon": np.arange(n_lon),
            },
        )
        .groupby(time.isel(sample=0).dt.year)
        .mean(dim=["time", "lat", "lon"])
    ).values
    np.testing.assert_allclose(
        annual_mean_series, expected_series, rtol=1e-4, atol=1e-7
    )
    logs = agg.get_logs(label="test")
    for source in ["target", "gen"]:
        r2 = logs[f"test/r2/a_{source}"]
        assert not np.isnan(r2)
        big_r2 = 50  # account for small denominators
        assert r2 > -big_r2  # can be -inf if NaNs improperly handled
        assert r2 < 1


@pytest.mark.parametrize("use_mock_distributed", [False, True])
def test__get_gathered_means(use_mock_distributed):
    """Test the private _get_gathered_means method, rather than the public
    get_logs method, because the public method returns some data converted to images,
    and so that portion can't easily be tested.
    """
    n_lat = 16
    n_lon = 32
    n_sample = 2
    n_time = 365 * 4 * 2  # two years, approximately
    area_weights = torch.ones(n_lat, n_lon).to(fme.get_device())
    agg = PairedGlobalMeanAnnualAggregator(
        ops=LatLonOperations(area_weights),
        timestep=TIMESTEP,
    )
    target_data = {
        "a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())
    }
    gen_data = {"a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())}
    time = xr.DataArray(
        [
            [
                (
                    cftime.DatetimeProlepticGregorian(
                        2000 + j, 1, 1
                    )  # add j so that years are not the same
                    + i * datetime.timedelta(hours=6)
                )
                for i in range(n_time)
            ]
            for j in range(n_sample)
        ],
        dims=["sample", "time"],
    )
    agg.record_batch(time, target_data, gen_data)
    if use_mock_distributed:
        world_size = 2
        with mock_distributed(world_size=world_size):
            result = agg._get_gathered_means()
            assert result is not None
            target, gen = result
            combined = agg.get_dataset()
    else:
        world_size = 1
        result = agg._get_gathered_means()
        assert result is not None
        target, gen = result
        combined = agg.get_dataset()
    for dataset in (target, gen, combined):
        assert set(dataset.dims).issuperset({"sample", "year"})
        assert list(dataset.year.values) == [2000, 2001, 2002]
        assert dataset.sizes["sample"] == n_sample * world_size
    assert set(combined.coords["source"].values) == set(["target", "prediction"])


def test_annual_aggregator():
    n_lat = 4
    n_lon = 8
    # need to have two actual full years of data for plotting to get exercised
    n_sample = 2
    n_time = 365 * 4 * 2
    area_weights = torch.ones(n_lat, n_lon).to(fme.get_device())
    agg = GlobalMeanAnnualAggregator(
        ops=LatLonOperations(area_weights), timestep=TIMESTEP
    )
    data = {"a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())}
    time = xr.DataArray(
        [
            [
                (
                    cftime.DatetimeProlepticGregorian(2000, 1, 1)
                    + i * datetime.timedelta(hours=6)
                )
                for i in range(n_time)
            ]
            for _ in range(n_sample)
        ],
        dims=["sample", "time"],
    )
    agg.record_batch(time, data)
    logs = agg.get_logs(label="test")
    assert len(logs) > 0
    assert "test/a" in logs
    assert isinstance(logs["test/a"], plt.Figure)


def test_annual_aggregator_with_nans():
    torch.manual_seed(0)
    n_lat = 4
    n_lon = 8
    # need to have two actual full years of data for plotting to get exercised
    n_sample = 2
    n_time = 365 * 4 * 2
    area_weights = torch.ones(n_lat, n_lon).to(fme.get_device())

    data = {"a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())}
    data["a"][:, :, 1, 1] = float("nan")
    masks = {"mask_a": torch.ones_like(data["a"][0, 0])}
    masks["mask_a"][1, 1] = 0
    mask_provider = MaskProvider(masks).to(get_device())

    agg = GlobalMeanAnnualAggregator(
        ops=LatLonOperations(area_weights, mask_provider), timestep=TIMESTEP
    )
    time = xr.DataArray(
        [
            [
                (
                    cftime.DatetimeProlepticGregorian(2000, 1, 1)
                    + i * datetime.timedelta(hours=6)
                )
                for i in range(n_time)
            ]
            for _ in range(n_sample)
        ],
        dims=["sample", "time"],
    )
    agg.record_batch(time, data)
    expected_series = (
        xr.DataArray(
            data["a"].cpu().numpy(),
            dims={
                "sample": np.arange(n_sample),
                "time": time,
                "lat": np.arange(n_lat),
                "lon": np.arange(n_lon),
            },
        )
        .groupby(time.isel(sample=0).dt.year)
        .mean(dim=["time", "lat", "lon"])
    ).values
    annual_mean_series = agg.get_dataset()["a"].values
    np.testing.assert_allclose(
        annual_mean_series, expected_series, rtol=1e-4, atol=1e-7
    )



import abc
from collections.abc import Iterable, Sized
from typing import Generic, Protocol, TypeVar

T = TypeVar("T", covariant=True)


class DataLoader(Protocol, Generic[T], Sized, Iterable[T]):
    pass


PS = TypeVar("PS")  # prognostic state
FD = TypeVar("FD", covariant=True)  # forcing data


class InferenceDataABC(abc.ABC, Generic[PS, FD]):
    @property
    @abc.abstractmethod
    def initial_condition(self) -> PS: ...

    @property
    @abc.abstractmethod
    def loader(self) -> DataLoader[FD]: ...


class SimpleInferenceData(InferenceDataABC[PS, FD]):
    def __init__(
        self,
        initial_condition: PS,
        loader: DataLoader[FD],
    ):
        self._initial_condition = initial_condition
        self._loader = loader

    @property
    def initial_condition(self) -> PS:
        return self._initial_condition

    @property
    def loader(self) -> DataLoader[FD]:
        return self._loader


class GriddedDataABC(abc.ABC, Generic[T]):
    @property
    @abc.abstractmethod
    def loader(self) -> DataLoader[T]: ...

    @property
    @abc.abstractmethod
    def n_samples(self) -> int: ...

    @property
    @abc.abstractmethod
    def n_batches(self) -> int: ...

    @property
    @abc.abstractmethod
    def batch_size(self) -> int: ...

    @abc.abstractmethod
    def set_epoch(self, epoch: int): ...

    @abc.abstractmethod
    def log_info(self, name: str):
        """
        Report information about the data using logging.info.
        """
        ...



# flake8: noqa
# Copied from tag v0.1.0 https://github.com/NVIDIA/modulus-makani/tree/0218658e925a3c88c6513022a33ecc1647735c2e
# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

# import FactorizedTensor from tensorly for tensorized operations
import tensorly as tl
import torch
import torch.nn as nn

tl.set_backend("pytorch")

# import convenience functions for factorized tensors
from tltorch.factorized_tensors.core import FactorizedTensor

# for the experimental module
from .contractions import _contract_rank
from .factorizations import get_contract_fun


class SpectralConv(nn.Module):
    """
    Spectral Convolution implemented via SHT or FFT. Designed for convolutions on the two-sphere S2
    using the Spherical Harmonic Transforms in torch-harmonics, but supports convolutions on the periodic
    domain via the RealFFT2 and InverseRealFFT2 wrappers.
    """

    def __init__(
        self,
        forward_transform,
        inverse_transform,
        in_channels,
        out_channels,
        operator_type="diagonal",
        separable=False,
        bias=False,
        gain=1.0,
    ):
        super(SpectralConv, self).__init__()

        self.forward_transform = forward_transform
        self.inverse_transform = inverse_transform

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.modes_lat = self.inverse_transform.lmax
        self.modes_lon = self.inverse_transform.mmax

        self.scale_residual = (
            self.forward_transform.nlat != self.inverse_transform.nlat
        ) or (self.forward_transform.nlon != self.inverse_transform.nlon)
        if hasattr(self.forward_transform, "grid"):
            self.scale_residual = self.scale_residual or (
                self.forward_transform.grid != self.inverse_transform.grid
            )

        # remember factorization details
        self.operator_type = operator_type
        self.separable = separable

        assert self.inverse_transform.lmax == self.modes_lat
        assert self.inverse_transform.mmax == self.modes_lon

        weight_shape = [in_channels]

        if not self.separable:
            weight_shape += [out_channels]

        self.modes_lat_local = self.modes_lat
        self.modes_lon_local = self.modes_lon
        self.nlat_local = self.inverse_transform.nlat
        self.nlon_local = self.inverse_transform.nlon

        # unpadded weights
        if self.operator_type == "diagonal":
            weight_shape += [self.modes_lat_local, self.modes_lon_local]
        elif self.operator_type == "dhconv":
            weight_shape += [self.modes_lat_local]
        else:
            raise ValueError(f"Unsupported operator type f{self.operator_type}")

        # Compute scaling factor for correct initialization
        scale = math.sqrt(gain / in_channels) * torch.ones(
            self.modes_lat_local, dtype=torch.complex64
        )
        # seemingly the first weight is not really complex, so we need to account for that
        scale[0] *= math.sqrt(2.0)
        init = scale * torch.randn(*weight_shape, dtype=torch.complex64)
        self.weight = nn.Parameter(torch.view_as_real(init))

        if self.operator_type == "dhconv":
            self.weight.is_shared_mp = ["matmul", "w"]
            self.weight.sharded_dims_mp = [None for _ in weight_shape]
            self.weight.sharded_dims_mp[-1] = "h"
        else:
            self.weight.is_shared_mp = ["matmul"]
            self.weight.sharded_dims_mp = [None for _ in weight_shape]
            self.weight.sharded_dims_mp[-1] = "w"
            self.weight.sharded_dims_mp[-2] = "h"

        # get the contraction handle. This should return a pyTorch contraction
        self._contract = get_contract_fun(
            self.weight,
            implementation="factorized",
            separable=separable,
            complex=True,
            operator_type=operator_type,
        )

        if bias == "constant":
            self.bias = nn.Parameter(torch.zeros(1, self.out_channels, 1, 1))
        elif bias == "position":
            self.bias = nn.Parameter(
                torch.zeros(1, self.out_channels, self.nlat_local, self.nlon_local)
            )
            self.bias.is_shared_mp = ["matmul"]
            self.bias.sharded_dims_mp = [None, None, "h", "w"]

    def forward(self, x):
        dtype = x.dtype
        residual = x
        x = x.float()
        B, C, H, W = x.shape

        with torch.amp.autocast("cuda", enabled=False):
            x = self.forward_transform(x).contiguous()
            if self.scale_residual:
                residual = self.inverse_transform(x)
                residual = residual.to(dtype)

        # approach with unpadded weights
        weight = torch.view_as_complex(self.weight)
        xp = self._contract(
            x, weight, separable=self.separable, operator_type=self.operator_type
        )
        x = xp.contiguous()

        with torch.amp.autocast("cuda", enabled=False):
            x = self.inverse_transform(x)

        if hasattr(self, "bias"):
            x = x + self.bias

        x = x.to(dtype=dtype)

        return x, residual


class FactorizedSpectralConv(nn.Module):
    """
    Factorized version of SpectralConv. Uses tensorly-torch to keep the weights factorized
    """

    def __init__(
        self,
        forward_transform,
        inverse_transform,
        in_channels,
        out_channels,
        operator_type="diagonal",
        rank=0.2,
        factorization=None,
        separable=False,
        decomposition_kwargs=dict(),
        bias=False,
        gain=1.0,
    ):
        super(FactorizedSpectralConv, self).__init__()

        self.forward_transform = forward_transform
        self.inverse_transform = inverse_transform

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.modes_lat = self.inverse_transform.lmax
        self.modes_lon = self.inverse_transform.mmax

        self.scale_residual = (
            self.forward_transform.nlat != self.inverse_transform.nlat
        ) or (self.forward_transform.nlon != self.inverse_transform.nlon)
        if hasattr(self.forward_transform, "grid"):
            self.scale_residual = self.scale_residual or (
                self.forward_transform.grid != self.inverse_transform.grid
            )

        # Make sure we are using a Complex Factorized Tensor
        if factorization is None:
            factorization = "ComplexDense"  # No factorization
        complex_weight = factorization[:7].lower() == "complex"

        # remember factorization details
        self.operator_type = operator_type
        self.rank = rank
        self.factorization = factorization
        self.separable = separable

        assert self.inverse_transform.lmax == self.modes_lat
        assert self.inverse_transform.mmax == self.modes_lon

        weight_shape = [in_channels]

        if not self.separable:
            weight_shape += [out_channels]

        self.modes_lat_local = self.modes_lat
        self.modes_lon_local = self.modes_lon

        # unpadded weights
        if self.operator_type == "diagonal":
            weight_shape += [self.modes_lat_local, self.modes_lon_local]
        elif self.operator_type == "dhconv":
            weight_shape += [self.modes_lat_local]
        elif self.operator_type == "rank":
            weight_shape += [self.rank]
        else:
            raise ValueError(f"Unsupported operator type f{self.operator_type}")

        # form weight tensors
        self.weight = FactorizedTensor.new(
            weight_shape,
            rank=self.rank,
            factorization=factorization,
            fixed_rank_modes=False,
            **decomposition_kwargs,
        )
        # initialization of weights
        scale = math.sqrt(gain / float(weight_shape[0]))
        self.weight.normal_(mean=0.0, std=scale)

        # get the contraction handle
        if operator_type == "rank":
            self._contract = _contract_rank
        else:
            self._contract = get_contract_fun(
                self.weight,
                implementation="reconstructed",
                separable=separable,
                complex=complex_weight,
                operator_type=operator_type,
            )

        if bias == "constant":
            self.bias = nn.Parameter(torch.zeros(1, self.out_channels, 1, 1))
        elif bias == "position":
            self.bias = nn.Parameter(
                torch.zeros(1, self.out_channels, self.nlat_local, self.nlon_local)
            )
            self.bias.is_shared_mp = ["matmul"]
            self.bias.sharded_dims_mp = [None, None, "h", "w"]

    def forward(self, x):
        dtype = x.dtype
        residual = x
        x = x.float()

        with torch.amp.autocast("cuda", enabled=False):
            x = self.forward_transform(x).contiguous()
            if self.scale_residual:
                residual = self.inverse_transform(x)
                residual = residual.to(dtype)

        if self.operator_type == "rank":
            xp = self._contract(x, self.weight, self.lat_weight, self.lon_weight)
        else:
            xp = self._contract(
                x,
                self.weight,
                separable=self.separable,
                operator_type=self.operator_type,
            )
        x = xp.contiguous()

        with torch.amp.autocast("cuda", enabled=False):
            x = self.inverse_transform(x)

        if hasattr(self, "bias"):
            x = x + self.bias

        x = x.type(dtype)

        return x, residual
